"""Multiprocessing files for alignment functions in MFA"""
from __future__ import annotations

import multiprocessing as mp
import os
import re
import statistics
import subprocess
import sys
import time
import traceback
from queue import Empty
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from ..exceptions import AlignmentError, AlignmentExportError
from ..multiprocessing.helper import Stopped
from ..textgrid import (
    ctms_to_textgrids_non_mp,
    export_textgrid,
    generate_tiers,
    output_textgrid_writing_errors,
    parse_from_phone,
    parse_from_word,
    parse_from_word_no_cleanup,
    process_ctm_line,
)
from ..utils import thirdparty_binary
from .helper import run_mp, run_non_mp

if TYPE_CHECKING:
    from ..aligner.adapting import AdaptingAligner
    from ..aligner.base import BaseAligner
    from ..config.align_config import AlignConfig, ConfigDict
    from ..corpus.classes import (
        CleanupWordCtmArguments,
        CombineCtmArguments,
        ExportTextGridArguments,
        File,
        NoCleanupWordCtmArguments,
        PhoneCtmArguments,
        Utterance,
    )
    from ..textgrid import CtmInterval
    from ..trainers import BaseTrainer, LdaTrainer, MonophoneTrainer

    ConfigType = Union[BaseTrainer, AlignConfig]

    IterationType = Union[str, int]

    AlignerType = Union[BaseTrainer, BaseAligner]
    CtmType = List[CtmInterval]

CtmErrorDict = Dict[Tuple[str, int], str]

queue_polling_timeout = 1

__all__ = [
    "acc_stats",
    "align",
    "mono_align_equal",
    "tree_stats",
    "compile_train_graphs",
    "compile_information",
    "convert_alignments",
    "convert_ali_to_textgrids",
    "compute_alignment_improvement",
    "compare_alignments",
    "PhoneCtmProcessWorker",
    "CleanupWordCtmProcessWorker",
    "NoCleanupWordCtmProcessWorker",
    "CombineProcessWorker",
    "ExportPreparationProcessWorker",
    "ExportTextGridProcessWorker",
    "calc_fmllr",
    "calc_lda_mllt",
    "create_align_model",
    "ctms_to_textgrids_mp",
    "lda_acc_stats",
    "train_map",
    "parse_iteration_alignments",
]


def acc_stats_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ali_paths: Dict[str, str],
    acc_paths: Dict[str, str],
    model_path: str,
) -> None:
    """
    Multiprocessing function for accumulating stats in GMM training

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    acc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    model_path: str
        Path to the acoustic model file
    """
    model_path = model_path
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            acc_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-acc-stats-ali"),
                    model_path,
                    feature_strings[dict_name],
                    f"ark,s,cs:{ali_paths[dict_name]}",
                    acc_paths[dict_name],
                ],
                stderr=log_file,
                env=os.environ,
            )
            acc_proc.communicate()


def acc_stats(aligner: AlignerType):
    """
    Multiprocessing function that computes stats for GMM training

    See http://kaldi-asr.org/doc/gmm-acc-stats-ali_8cc.html for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_mono.sh
    for the bash script this function was extracted from

    Parameters
    ----------
    aligner : :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Trainer
    """
    arguments = [j.acc_stats_arguments(aligner) for j in aligner.corpus.jobs]

    if aligner.use_mp:
        run_mp(acc_stats_func, arguments, aligner.working_log_directory)
    else:
        run_non_mp(acc_stats_func, arguments, aligner.working_log_directory)

    log_path = os.path.join(aligner.working_log_directory, f"update.{aligner.iteration}.log")
    with open(log_path, "w") as log_file:
        acc_files = []
        for a in arguments:
            acc_files.extend(a.acc_paths.values())
        sum_proc = subprocess.Popen(
            [thirdparty_binary("gmm-sum-accs"), "-"] + acc_files,
            stdout=subprocess.PIPE,
            stderr=log_file,
            env=os.environ,
        )
        est_proc = subprocess.Popen(
            [
                thirdparty_binary("gmm-est"),
                f"--write-occs={aligner.next_occs_path}",
                f"--mix-up={aligner.current_gaussians}",
                f"--power={aligner.power}",
                aligner.current_model_path,
                "-",
                aligner.next_model_path,
            ],
            stdin=sum_proc.stdout,
            stderr=log_file,
            env=os.environ,
        )
        est_proc.communicate()
    avg_like_pattern = re.compile(
        r"Overall avg like per frame \(Gaussian only\) = (?P<like>[-.,\d]+) over (?P<frames>[.\d+e]) frames"
    )
    average_logdet_pattern = re.compile(
        r"Overall average logdet is (?P<logdet>[-.,\d]+) over (?P<frames>[.\d+e]) frames"
    )
    avg_like_sum = 0
    avg_like_frames = 0
    average_logdet_sum = 0
    average_logdet_frames = 0
    for a in arguments:
        with open(a.log_path, "r", encoding="utf8") as f:
            for line in f:
                m = re.search(avg_like_pattern, line)
                if m:
                    like = float(m.group("like"))
                    frames = float(m.group("frames"))
                    avg_like_sum += like * frames
                    avg_like_frames += frames
                m = re.search(average_logdet_pattern, line)
                if m:
                    logdet = float(m.group("logdet"))
                    frames = float(m.group("frames"))
                    average_logdet_sum += logdet * frames
                    average_logdet_frames += frames
    if avg_like_frames:
        log_like = avg_like_sum / avg_like_frames
        if average_logdet_frames:
            log_like += average_logdet_sum / average_logdet_frames
        aligner.logger.debug(f"Likelihood for iteration {aligner.iteration}: {log_like}")

    if not aligner.debug:
        for f in acc_files:
            os.remove(f)


def compile_train_graphs_func(
    log_path: str,
    dictionaries: List[str],
    tree_path: str,
    model_path: str,
    text_int_paths: Dict[str, str],
    disambig_paths: Dict[str, str],
    lexicon_fst_paths: Dict[str, str],
    fst_scp_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function to compile training graphs

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    tree_path: str
        Path to the acoustic model tree file
    model_path: str
        Path to the acoustic model file
    text_int_paths: Dict[str, str]
        Dictionary of text int files per dictionary name
    disambig_paths: Dict[str, str]
        Dictionary of disambiguation symbol int files per dictionary name
    lexicon_fst_paths: Dict[str, str]
        Dictionary of L.fst files per dictionary name
    fst_scp_paths: Dict[str, str]
        Dictionary of utterance FST scp files per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            disambig_path = disambig_paths[dict_name]
            fst_scp_path = fst_scp_paths[dict_name]
            fst_ark_path = fst_scp_path.replace(".scp", ".ark")
            text_path = text_int_paths[dict_name]
            proc = subprocess.Popen(
                [
                    thirdparty_binary("compile-train-graphs"),
                    f"--read-disambig-syms={disambig_path}",
                    tree_path,
                    model_path,
                    lexicon_fst_paths[dict_name],
                    f"ark:{text_path}",
                    f"ark,scp:{fst_ark_path},{fst_scp_path}",
                ],
                stderr=log_file,
                env=os.environ,
            )
            proc.communicate()


def compile_train_graphs(aligner: AlignerType) -> None:
    """
    Multiprocessing function that compiles training graphs for utterances

    See http://kaldi-asr.org/doc/compile-train-graphs_8cc.html for more details
    on the Kaldi binary this function calls.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_mono.sh
    for the bash script that this function was extracted from.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    aligner.logger.debug("Compiling training graphs...")
    begin = time.time()
    log_directory = aligner.working_log_directory
    os.makedirs(log_directory, exist_ok=True)
    jobs = [x.compile_train_graph_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(compile_train_graphs_func, jobs, log_directory)
    else:
        run_non_mp(compile_train_graphs_func, jobs, log_directory)
    aligner.logger.debug(f"Compiling training graphs took {time.time() - begin}")


def mono_align_equal_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    fst_scp_paths: Dict[str, str],
    ali_ark_paths: Dict[str, str],
    acc_paths: Dict[str, str],
    model_path: str,
):
    """
    Multiprocessing function for initializing monophone alignments

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    fst_scp_paths: Dict[str, str]
        Dictionary of utterance FST scp files per dictionary name
    ali_ark_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    acc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    model_path: str
        Path to the acoustic model file
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            fst_path = fst_scp_paths[dict_name]
            ali_path = ali_ark_paths[dict_name]
            acc_path = acc_paths[dict_name]
            align_proc = subprocess.Popen(
                [
                    thirdparty_binary("align-equal-compiled"),
                    f"scp:{fst_path}",
                    feature_strings[dict_name],
                    f"ark:{ali_path}",
                ],
                stderr=log_file,
                env=os.environ,
            )
            align_proc.communicate()
            stats_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-acc-stats-ali"),
                    "--binary=true",
                    model_path,
                    feature_strings[dict_name],
                    f"ark:{ali_path}",
                    acc_path,
                ],
                stdin=align_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            stats_proc.communicate()


def mono_align_equal(aligner: MonophoneTrainer):
    """
    Multiprocessing function that creates equal alignments for base monophone training

    See http://kaldi-asr.org/doc/align-equal-compiled_8cc.html for more details
    on the Kaldi binary this function calls.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_mono.sh
    for the bash script that this function was extracted from.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.MonophoneTrainer`
        Monophone trainer
    """

    arguments = [x.mono_align_equal_arguments(aligner) for x in aligner.corpus.jobs]

    if aligner.use_mp:
        run_mp(mono_align_equal_func, arguments, aligner.log_directory)
    else:
        run_non_mp(mono_align_equal_func, arguments, aligner.log_directory)

    log_path = os.path.join(aligner.working_log_directory, "update.0.log")
    with open(log_path, "w") as log_file:
        acc_files = []
        for x in arguments:
            acc_files.extend(sorted(x.acc_paths.values()))
        sum_proc = subprocess.Popen(
            [thirdparty_binary("gmm-sum-accs"), "-"] + acc_files,
            stderr=log_file,
            stdout=subprocess.PIPE,
            env=os.environ,
        )
        est_proc = subprocess.Popen(
            [
                thirdparty_binary("gmm-est"),
                "--min-gaussian-occupancy=3",
                f"--mix-up={aligner.current_gaussians}",
                f"--power={aligner.power}",
                aligner.current_model_path,
                "-",
                aligner.next_model_path,
            ],
            stderr=log_file,
            stdin=sum_proc.stdout,
            env=os.environ,
        )
        est_proc.communicate()
    if not aligner.debug:
        for f in acc_files:
            os.remove(f)


def align_func(
    log_path: str,
    dictionaries: List[str],
    fst_scp_paths: Dict[str, str],
    feature_strings: Dict[str, str],
    model_path: str,
    ali_paths: Dict[str, str],
    score_paths: Dict[str, str],
    loglike_paths: Dict[str, str],
    align_options: ConfigDict,
):
    """
    Multiprocessing function for alignment

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    fst_scp_paths: Dict[str, str]
        Dictionary of FST scp file paths per dictionary name
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    model_path: str
        Path to the acoustic model file
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    score_paths: Dict[str, str]
        Dictionary of scores files per dictionary name
    loglike_paths: Dict[str, str]
        Dictionary of log likelihood files per dictionary name
    align_options: ConfigDict
        Options for alignment
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            fst_path = fst_scp_paths[dict_name]
            ali_path = ali_paths[dict_name]
            com = [
                thirdparty_binary("gmm-align-compiled"),
                f"--transition-scale={align_options['transition_scale']}",
                f"--acoustic-scale={align_options['acoustic_scale']}",
                f"--self-loop-scale={align_options['self_loop_scale']}",
                f"--beam={align_options['beam']}",
                f"--retry-beam={align_options['retry_beam']}",
                "--careful=false",
                "-",
                f"scp:{fst_path}",
                feature_string,
                f"ark:{ali_path}",
            ]
            if align_options["debug"]:
                loglike_path = loglike_paths[dict_name]
                score_path = score_paths[dict_name]
                com.insert(1, f"--write-per-frame-acoustic-loglikes=ark,t:{loglike_path}")
                com.append(f"ark,t:{score_path}")

            boost_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-boost-silence"),
                    f"--boost={align_options['boost_silence']}",
                    align_options["optional_silence_csl"],
                    model_path,
                    "-",
                ],
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            align_proc = subprocess.Popen(
                com, stderr=log_file, stdin=boost_proc.stdout, env=os.environ
            )
            align_proc.communicate()


def align(aligner: AlignerType) -> None:
    """
    Multiprocessing function that aligns based on the current model

    See http://kaldi-asr.org/doc/gmm-align-compiled_8cc.html and
    http://kaldi-asr.org/doc/gmm-boost-silence_8cc.html for more details
    on the Kaldi binary this function calls.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/align_si.sh
    for the bash script this function was based on.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    begin = time.time()
    log_directory = aligner.working_log_directory

    arguments = [x.align_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(align_func, arguments, log_directory)
    else:
        run_non_mp(align_func, arguments, log_directory)

    error_logs = []
    for j in arguments:

        with open(j.log_path, "r", encoding="utf8") as f:
            for line in f:
                if line.strip().startswith("ERROR"):
                    error_logs.append(j.log_path)
                    break
    if error_logs:
        raise AlignmentError(error_logs)
    aligner.logger.debug(f"Alignment round took {time.time() - begin}")


def compile_information_func(align_log_path: str) -> Dict[str, Union[List[str], float, int]]:
    """
    Multiprocessing function for compiling information about alignment

    Parameters
    ----------
    align_log_path: str
        Log path for alignment

    Returns
    -------
    Dict
        Information about log-likelihood and number of unaligned files
    """
    average_logdet_pattern = re.compile(
        r"Overall average logdet is (?P<logdet>[-.,\d]+) over (?P<frames>[.\d+e]+) frames"
    )
    log_like_pattern = re.compile(
        r"^LOG .* Overall log-likelihood per frame is (?P<log_like>[-0-9.]+) over (?P<frames>\d+) frames.*$"
    )

    decode_error_pattern = re.compile(
        r"^WARNING .* Did not successfully decode file (?P<utt>.*?), .*$"
    )

    data = {"unaligned": [], "too_short": [], "log_like": 0, "total_frames": 0}
    with open(align_log_path, "r", encoding="utf8") as f:
        for line in f:
            decode_error_match = re.match(decode_error_pattern, line)
            if decode_error_match:
                data["unaligned"].append(decode_error_match.group("utt"))
                continue
            log_like_match = re.match(log_like_pattern, line)
            if log_like_match:
                log_like = log_like_match.group("log_like")
                frames = log_like_match.group("frames")
                data["log_like"] = float(log_like)
                data["total_frames"] = int(frames)
            m = re.search(average_logdet_pattern, line)
            if m:
                logdet = float(m.group("logdet"))
                frames = float(m.group("frames"))
                data["logdet"] = logdet
                data["logdet_frames"] = frames
    return data


def compile_information(aligner: AlignerType) -> Tuple[Dict[str, str], float]:
    """
    Compiles information about alignment, namely what the overall log-likelihood was
    and how many files were unaligned

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner

    Returns
    -------
    Dict
        Unaligned files
    float
        Log-likelihood of alignment
    """
    compile_info_begin = time.time()

    jobs = [x.compile_information_arguments(aligner) for x in aligner.corpus.jobs]

    if aligner.use_mp:
        alignment_info = run_mp(
            compile_information_func, jobs, aligner.working_log_directory, True
        )
    else:
        alignment_info = run_non_mp(
            compile_information_func, jobs, aligner.working_log_directory, True
        )

    unaligned = {}
    avg_like_sum = 0
    avg_like_frames = 0
    average_logdet_sum = 0
    average_logdet_frames = 0
    for data in alignment_info.values():
        avg_like_frames += data["total_frames"]
        avg_like_sum += data["log_like"] * data["total_frames"]
        if "logdet_frames" in data:
            average_logdet_frames += data["logdet_frames"]
            average_logdet_sum += data["logdet"] * data["logdet_frames"]
        for u in data["unaligned"]:
            unaligned[u] = "Beam too narrow"
        for u in data["too_short"]:
            unaligned[u] = "Segment too short"

    if not avg_like_frames:
        aligner.logger.warning(
            "No files were aligned, this likely indicates serious problems with the aligner."
        )
    aligner.logger.debug(f"Compiling information took {time.time() - compile_info_begin}")
    log_like = avg_like_sum / avg_like_frames
    if average_logdet_sum:
        log_like += average_logdet_sum / average_logdet_frames
    return unaligned, log_like


def compute_alignment_improvement_func(
    log_path: str,
    dictionaries: List[str],
    model_path: str,
    text_int_paths: Dict[str, str],
    word_boundary_paths: Dict[str, str],
    ali_paths: Dict[str, str],
    frame_shift: int,
    reversed_phone_mappings: Dict[str, Dict[int, str]],
    positions: Dict[str, List[str]],
    phone_ctm_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function for computing alignment improvement over training

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    model_path: str
        Path to the acoustic model file
    text_int_paths: Dict[str, str]
        Dictionary of text int files per dictionary name
    word_boundary_paths: Dict[str, str]
        Dictionary of word boundary files per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    frame_shift: int
        Frame shift of feature generation, in ms
    reversed_phone_mappings: Dict[str, Dict[int, str]]
        Mapping of phone IDs to phone labels per dictionary name
    positions: Dict[str, List[str]]
        Positions per dictionary name
    phone_ctm_paths: Dict[str, str]
        Dictionary of phone ctm files per dictionary name
    """
    try:

        frame_shift = frame_shift / 1000
        with open(log_path, "w", encoding="utf8") as log_file:
            for dict_name in dictionaries:
                text_int_path = text_int_paths[dict_name]
                ali_path = ali_paths[dict_name]
                phone_ctm_path = phone_ctm_paths[dict_name]
                word_boundary_path = word_boundary_paths[dict_name]
                if os.path.exists(phone_ctm_path):
                    continue

                lin_proc = subprocess.Popen(
                    [
                        thirdparty_binary("linear-to-nbest"),
                        f"ark:{ali_path}",
                        f"ark:{text_int_path}",
                        "",
                        "",
                        "ark:-",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                det_proc = subprocess.Popen(
                    [thirdparty_binary("lattice-determinize-pruned"), "ark:-", "ark:-"],
                    stdin=lin_proc.stdout,
                    stderr=log_file,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                align_proc = subprocess.Popen(
                    [
                        thirdparty_binary("lattice-align-words"),
                        word_boundary_path,
                        model_path,
                        "ark:-",
                        "ark:-",
                    ],
                    stdin=det_proc.stdout,
                    stderr=log_file,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                phone_proc = subprocess.Popen(
                    [thirdparty_binary("lattice-to-phone-lattice"), model_path, "ark:-", "ark:-"],
                    stdin=align_proc.stdout,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                nbest_proc = subprocess.Popen(
                    [
                        thirdparty_binary("nbest-to-ctm"),
                        f"--frame-shift={frame_shift}",
                        "ark:-",
                        phone_ctm_path,
                    ],
                    stdin=phone_proc.stdout,
                    stderr=log_file,
                    env=os.environ,
                )
                nbest_proc.communicate()
                mapping = reversed_phone_mappings[dict_name]
                actual_lines = []
                with open(phone_ctm_path, "r", encoding="utf8") as f:
                    for line in f:
                        line = line.strip()
                        if line == "":
                            continue
                        line = line.split(" ")
                        utt = line[0]
                        begin = float(line[2])
                        duration = float(line[3])
                        end = begin + duration
                        label = line[4]
                        try:
                            label = mapping[int(label)]
                        except KeyError:
                            pass
                        for p in positions[dict_name]:
                            if label.endswith(p):
                                label = label[: -1 * len(p)]
                        actual_lines.append([utt, begin, end, label])
                with open(phone_ctm_path, "w", encoding="utf8") as f:
                    for line in actual_lines:
                        f.write(f"{' '.join(map(str, line))}\n")
    except Exception as e:
        raise (Exception(str(e)))


def parse_iteration_alignments(
    aligner: AlignerType, iteration: Optional[IterationType] = None
) -> Dict[str, List[Tuple[float, float, str]]]:
    """
    Function to parse phone CTMs in a given iteration

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    iteration: IterationType
        Iteration to compute over
    Returns
    -------
    Dict
        Per utterance CtmIntervals
    """
    if iteration is None:
        iteration = aligner.iteration
    data = {}
    for j in aligner.corpus.jobs:
        phone_ctm_path = os.path.join(aligner.working_directory, f"phone.{iteration}.{j.name}.ctm")
        with open(phone_ctm_path, "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if line == "":
                    continue
                line = line.split(" ")
                utt = line[0]
                begin = float(line[1])
                end = float(line[2])
                label = line[3]
                if utt not in data:
                    data[utt] = []
                data[utt].append((begin, end, label))
    return data


def compare_alignments(
    alignments_one: Dict[str, List[Tuple[float, float, str]]],
    alignments_two: Dict[str, List[Tuple[float, float, str]]],
    frame_shift: int,
) -> Tuple[int, Optional[float]]:
    """
    Compares two sets of alignments for difference

    Parameters
    ----------
    alignments_one: Dict
        First set of alignments
    alignments_two: Dict
        Second set of alignments
    frame_shift: int
        Frame shift in feature generation, in ms

    Returns
    -------
    int
        Difference in number of aligned files
    float
        Mean boundary difference between the two alignments
    """
    utterances_aligned_diff = len(alignments_two) - len(alignments_one)
    utts_one = set(alignments_one.keys())
    utts_two = set(alignments_two.keys())
    common_utts = utts_one.intersection(utts_two)
    differences = []
    for u in common_utts:
        end = alignments_one[u][-1][1]
        t = 0
        one_alignment = alignments_one[u]
        two_alignment = alignments_two[u]
        difference = 0
        while t < end:
            one_label = None
            two_label = None
            for b, e, l in one_alignment:
                if t < b:
                    continue
                if t >= e:
                    break
                one_label = l
            for b, e, l in two_alignment:
                if t < b:
                    continue
                if t >= e:
                    break
                two_label = l
            if one_label != two_label:
                difference += frame_shift
            t += frame_shift
        difference /= end
        differences.append(difference)
    if differences:
        mean_difference = statistics.mean(differences)
    else:
        mean_difference = None
    return utterances_aligned_diff, mean_difference


def compute_alignment_improvement(aligner: AlignerType) -> None:
    """
    Computes aligner improvements in terms of number of aligned files and phone boundaries
    for debugging purposes

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    jobs = [x.alignment_improvement_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(compute_alignment_improvement_func, jobs, aligner.working_log_directory)
    else:
        run_non_mp(compute_alignment_improvement_func, jobs, aligner.working_log_directory)

    alignment_diff_path = os.path.join(aligner.working_directory, "train_change.csv")
    if aligner.iteration == 0 or aligner.iteration not in aligner.realignment_iterations:
        return
    ind = aligner.realignment_iterations.index(aligner.iteration)
    if ind != 0:
        previous_iteration = aligner.realignment_iterations[ind - 1]
    else:
        previous_iteration = 0
    try:
        previous_alignments = parse_iteration_alignments(aligner, previous_iteration)
    except FileNotFoundError:
        return
    current_alignments = parse_iteration_alignments(aligner)
    utterance_aligned_diff, mean_difference = compare_alignments(
        previous_alignments, current_alignments, aligner.feature_config.frame_shift
    )
    if not os.path.exists(alignment_diff_path):
        with open(alignment_diff_path, "w", encoding="utf8") as f:
            f.write(
                "iteration,number_aligned,number_previously_aligned,"
                "difference_in_utts_aligned,mean_boundary_change\n"
            )
    if aligner.iteration in aligner.realignment_iterations:
        with open(alignment_diff_path, "a", encoding="utf8") as f:
            f.write(
                f"{aligner.iteration},{len(current_alignments)},{len(previous_alignments)},"
                f"{utterance_aligned_diff},{mean_difference}\n"
            )
    if not aligner.debug:
        for j in jobs:
            for p in j.phone_ctm_paths:
                os.remove(p)


def ali_to_ctm_func(
    log_path: str,
    dictionaries: List[str],
    ali_paths: Dict[str, str],
    text_int_paths: Dict[str, str],
    word_boundary_int_paths: Dict[str, str],
    frame_shift: float,
    model_path: str,
    ctm_paths: Dict[str, str],
    word_mode: bool,
) -> None:
    """
    Multiprocessing function to convert alignment archives into CTM files

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    text_int_paths: Dict[str, str]
        Dictionary of text int files per dictionary name
    word_boundary_int_paths: Dict[str, str]
        Dictionary of word boundary int files per dictionary name
    frame_shift: float
        Frame shift of feature generation in seconds
    model_path: str
        Path to the acoustic model file
    ctm_paths: Dict[str, str]
        Dictionary of CTM files per dictionary name
    word_mode: bool
        Flag for whether to parse words or phones
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            ali_path = ali_paths[dict_name]
            text_int_path = text_int_paths[dict_name]
            ctm_path = ctm_paths[dict_name]
            word_boundary_int_path = word_boundary_int_paths[dict_name]
            if os.path.exists(ctm_path):
                return
            lin_proc = subprocess.Popen(
                [
                    thirdparty_binary("linear-to-nbest"),
                    "ark:" + ali_path,
                    "ark:" + text_int_path,
                    "",
                    "",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            align_words_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-align-words"),
                    word_boundary_int_path,
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=lin_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            if word_mode:
                nbest_proc = subprocess.Popen(
                    [
                        thirdparty_binary("nbest-to-ctm"),
                        f"--frame-shift={frame_shift}",
                        "ark:-",
                        ctm_path,
                    ],
                    stderr=log_file,
                    stdin=align_words_proc.stdout,
                    env=os.environ,
                )
            else:
                phone_proc = subprocess.Popen(
                    [thirdparty_binary("lattice-to-phone-lattice"), model_path, "ark:-", "ark:-"],
                    stdout=subprocess.PIPE,
                    stdin=align_words_proc.stdout,
                    stderr=log_file,
                    env=os.environ,
                )
                nbest_proc = subprocess.Popen(
                    [
                        thirdparty_binary("nbest-to-ctm"),
                        f"--frame-shift={frame_shift}",
                        "ark:-",
                        ctm_path,
                    ],
                    stdin=phone_proc.stdout,
                    stderr=log_file,
                    env=os.environ,
                )
            nbest_proc.communicate()


class NoCleanupWordCtmProcessWorker(mp.Process):
    """
    Multiprocessing worker for loading word CTM files without any clean up

    Parameters
    ----------
    job_name: int
        Job name
    to_process_queue: :class:`~multiprocessing.Queue`
        Return queue of jobs for later workers to process
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for processing
    error_catching: CtmErrorDict
        Dictionary for storing errors encountered
    arguments: :class:`~montreal_forced_aligner.multiprocessing.classes.NoCleanupWordCtmArguments`
        Arguments to pass to the CTM processing function
    """

    def __init__(
        self,
        job_name: int,
        to_process_queue: mp.Queue,
        stopped: Stopped,
        error_catching: CtmErrorDict,
        arguments: NoCleanupWordCtmArguments,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.dictionaries = arguments.dictionaries
        self.ctm_paths = arguments.ctm_paths
        self.to_process_queue = to_process_queue
        self.stopped = stopped
        self.error_catching = error_catching

        # Corpus information
        self.utterances = arguments.utterances

        # Dictionary information
        self.dictionary_data = arguments.dictionary_data

    def run(self) -> None:
        """
        Run the word processing with no clean up
        """
        current_file_data = {}

        def process_current(cur_utt: Utterance, current_labels: CtmType):
            """Process current stack of intervals"""
            actual_labels = parse_from_word_no_cleanup(
                current_labels, self.dictionary_data[dict_name].reversed_words_mapping
            )
            current_file_data[cur_utt.name] = actual_labels

        def process_current_file(cur_file: str):
            """Process current file and add to return queue"""
            self.to_process_queue.put(("word", cur_file, current_file_data))

        cur_utt = None
        cur_file = None
        utt_begin = 0
        current_labels = []
        try:
            for dict_name in self.dictionaries:
                with open(self.ctm_paths[dict_name], "r") as word_file:
                    for line in word_file:
                        line = line.strip()
                        if not line:
                            continue
                        interval = process_ctm_line(line)
                        utt = interval.utterance
                        if cur_utt is None:
                            cur_utt = self.utterances[dict_name][utt]
                            utt_begin = cur_utt.begin
                            cur_file = cur_utt.file_name

                        if utt != cur_utt:
                            process_current(cur_utt, current_labels)
                            cur_utt = self.utterances[dict_name][utt]
                            file_name = cur_utt.file_name
                            if file_name != cur_file:
                                process_current_file(cur_file)
                                current_file_data = {}
                                cur_file = file_name
                            current_labels = []
                        if utt_begin:
                            interval.shift_times(utt_begin)
                        current_labels.append(interval)
                if current_labels:
                    process_current(cur_utt, current_labels)
                    process_current_file(cur_file)
        except Exception:
            self.stopped.stop()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.error_catching[("word", self.job_name)] = "\n".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            )


class CleanupWordCtmProcessWorker(mp.Process):
    """
    Multiprocessing worker for loading word CTM files with cleaning up MFA-internal modifications

    Parameters
    ----------
    job_name: int
        Job name
    to_process_queue: :class:`~multiprocessing.Queue`
        Return queue of jobs for later workers to process
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for processing
    error_catching: CtmErrorDict
        Dictionary for storing errors encountered
    arguments: :class:`~montreal_forced_aligner.multiprocessing.classes.CleanupWordCtmArguments`
        Arguments to pass to the CTM processing function
    """

    def __init__(
        self,
        job_name: int,
        to_process_queue: mp.Queue,
        stopped: Stopped,
        error_catching: CtmErrorDict,
        arguments: CleanupWordCtmArguments,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.dictionaries = arguments.dictionaries
        self.ctm_paths = arguments.ctm_paths
        self.to_process_queue = to_process_queue
        self.stopped = stopped
        self.error_catching = error_catching

        # Corpus information
        self.utterances = arguments.utterances

        # Dictionary information
        self.dictionary_data = arguments.dictionary_data

    def run(self) -> None:
        """
        Run the word processing with clean up
        """
        current_file_data = {}

        def process_current(cur_utt: Utterance, current_labels: CtmType) -> None:
            """Process current stack of intervals"""
            text = cur_utt.text.split()
            actual_labels = parse_from_word(current_labels, text, self.dictionary_data[dict_name])

            current_file_data[cur_utt.name] = actual_labels

        def process_current_file(cur_file: str) -> None:
            """Process current file and add to return queue"""
            self.to_process_queue.put(("word", cur_file, current_file_data))

        cur_utt = None
        cur_file = None
        utt_begin = 0
        current_labels = []
        try:
            for dict_name in self.dictionaries:
                ctm_path = self.ctm_paths[dict_name]
                with open(ctm_path, "r") as word_file:
                    for line in word_file:
                        line = line.strip()
                        if not line:
                            continue
                        interval = process_ctm_line(line)
                        utt = interval.utterance
                        if cur_utt is None:
                            cur_utt = self.utterances[dict_name][utt]
                            utt_begin = cur_utt.begin
                            cur_file = cur_utt.file_name

                        if utt != cur_utt:
                            process_current(cur_utt, current_labels)
                            cur_utt = self.utterances[dict_name][utt]
                            utt_begin = cur_utt.begin
                            file_name = cur_utt.file_name
                            if file_name != cur_file:
                                process_current_file(cur_file)
                                current_file_data = {}
                                cur_file = file_name
                            current_labels = []
                        if utt_begin:
                            interval.shift_times(utt_begin)
                        current_labels.append(interval)
                if current_labels:
                    process_current(cur_utt, current_labels)
                    process_current_file(cur_file)
        except Exception:
            self.stopped.stop()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.error_catching[("word", self.job_name)] = "\n".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            )


class PhoneCtmProcessWorker(mp.Process):
    """
    Multiprocessing worker for loading phone CTM files

    Parameters
    ----------
    job_name: int
        Job name
    to_process_queue: :class:`~multiprocessing.Queue`
        Return queue of jobs for later workers to process
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for processing
    error_catching: CtmErrorDict
        Dictionary for storing errors encountered
    arguments: :class:`~montreal_forced_aligner.multiprocessing.classes.PhoneCtmArguments`
        Arguments to pass to the CTM processing function
    """

    def __init__(
        self,
        job_name: int,
        to_process_queue: mp.Queue,
        stopped: Stopped,
        error_catching: CtmErrorDict,
        arguments: PhoneCtmArguments,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.dictionaries = arguments.dictionaries
        self.ctm_paths = arguments.ctm_paths
        self.to_process_queue = to_process_queue
        self.stopped = stopped
        self.error_catching = error_catching

        self.utterances = arguments.utterances

        self.reversed_phone_mappings = arguments.reversed_phone_mappings
        self.positions = arguments.positions

    def run(self) -> None:
        """Run the phone processing"""
        cur_utt = None
        cur_file = None
        utt_begin = 0
        current_labels = []

        current_file_data = {}

        def process_current_utt(cur_utt: Utterance, current_labels: CtmType) -> None:
            """Process current stack of intervals"""
            actual_labels = parse_from_phone(
                current_labels, self.reversed_phone_mappings[dict_name], self.positions[dict_name]
            )
            current_file_data[cur_utt.name] = actual_labels

        def process_current_file(cur_file: str) -> None:
            """Process current file and add to return queue"""
            self.to_process_queue.put(("phone", cur_file, current_file_data))

        try:
            for dict_name in self.dictionaries:
                with open(self.ctm_paths[dict_name], "r") as word_file:
                    for line in word_file:
                        line = line.strip()
                        if not line:
                            continue
                        interval = process_ctm_line(line)
                        utt = interval.utterance
                        if cur_utt is None:
                            cur_utt = self.utterances[dict_name][utt]
                            cur_file = cur_utt.file_name
                            utt_begin = cur_utt.begin

                        if utt != cur_utt:

                            process_current_utt(cur_utt, current_labels)

                            cur_utt = self.utterances[dict_name][utt]
                            file_name = cur_utt.file_name
                            utt_begin = cur_utt.begin

                            if file_name != cur_file:
                                process_current_file(cur_file)
                                current_file_data = {}
                                cur_file = file_name
                            current_labels = []
                        if utt_begin:
                            interval.shift_times(utt_begin)
                        current_labels.append(interval)
                if current_labels:
                    process_current_utt(cur_utt, current_labels)
                    process_current_file(cur_file)
        except Exception:
            self.stopped.stop()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.error_catching[("phone", self.job_name)] = "\n".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            )


class CombineProcessWorker(mp.Process):
    """
    Multiprocessing worker for loading phone CTM files

    Parameters
    ----------
    job_name: int
        Job name
    to_process_queue: :class:`~multiprocessing.Queue`
        Input queue of phone and word ctms to combine
    to_export_queue: :class:`~multiprocessing.Queue`
        Export queue of combined CTMs
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for processing
    finished_combining: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Signal that this worker has finished combining all CTMs
    error_catching: CtmErrorDict
        Dictionary for storing errors encountered
    arguments: :class:`~montreal_forced_aligner.multiprocessing.classes.CombineCtmArguments`
        Arguments to pass to the CTM combining function
    """

    def __init__(
        self,
        job_name: int,
        to_process_queue: mp.Queue,
        to_export_queue: mp.Queue,
        stopped: Stopped,
        finished_combining: Stopped,
        error_catching: CtmErrorDict,
        arguments: CombineCtmArguments,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.to_process_queue = to_process_queue
        self.to_export_queue = to_export_queue
        self.stopped = stopped
        self.finished_combining = finished_combining
        self.error_catching = error_catching

        self.files = arguments.files
        self.dictionary_data = arguments.dictionary_data
        self.cleanup_textgrids = arguments.cleanup_textgrids

    def run(self) -> None:
        """Run the combination function"""
        sum_time = 0
        count_time = 0
        phone_data = {}
        word_data = {}
        while True:
            try:
                w_p, file_name, data = self.to_process_queue.get(timeout=queue_polling_timeout)
                begin_time = time.time()
            except Empty:
                if self.finished_combining.stop_check():
                    break
                continue
            self.to_process_queue.task_done()
            if self.stopped.stop_check():
                continue
            if w_p == "phone":
                if file_name in word_data:
                    word_ctm = word_data.pop(file_name)
                    phone_ctm = data
                else:
                    phone_data[file_name] = data
                    continue
            else:
                if file_name in phone_data:
                    phone_ctm = phone_data.pop(file_name)
                    word_ctm = data
                else:
                    word_data[file_name] = data
                    continue
            try:
                file = self.files[file_name]
                for u_name, u in file.utterances.items():
                    if u_name not in word_ctm:
                        continue
                    u.word_labels = word_ctm[u_name]
                    u.phone_labels = phone_ctm[u_name]
                data = generate_tiers(file, cleanup_textgrids=self.cleanup_textgrids)
                self.to_export_queue.put((file_name, data))
            except Exception:
                self.stopped.stop()
                exc_type, exc_value, exc_traceback = sys.exc_info()
                self.error_catching[("combining", self.job_name)] = "\n".join(
                    traceback.format_exception(exc_type, exc_value, exc_traceback)
                )

            sum_time += time.time() - begin_time
            count_time += 1


class ExportTextGridProcessWorker(mp.Process):
    """
    Multiprocessing worker for exporting TextGrids

    Parameters
    ----------
    for_write_queue: :class:`~multiprocessing.Queue`
        Input queue of files to export
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for processing
    finished_processing: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Input signal that all jobs have been added and no more new ones will come in
    textgrid_errors: CtmErrorDict
        Dictionary for storing errors encountered
    arguments: :class:`~montreal_forced_aligner.multiprocessing.classes.ExportTextGridArguments`
        Arguments to pass to the TextGrid export function
    """

    def __init__(
        self,
        for_write_queue: mp.Queue,
        stopped: Stopped,
        finished_processing: Stopped,
        textgrid_errors: Dict[str, str],
        arguments: ExportTextGridArguments,
    ):
        mp.Process.__init__(self)
        self.for_write_queue = for_write_queue
        self.stopped = stopped
        self.finished_processing = finished_processing
        self.textgrid_errors = textgrid_errors

        self.files = arguments.files
        self.output_directory = arguments.output_directory
        self.backup_output_directory = arguments.backup_output_directory

        self.frame_shift = arguments.frame_shift

    def run(self) -> None:
        """Run the exporter function"""
        while True:
            try:
                file_name, data = self.for_write_queue.get(timeout=queue_polling_timeout)
            except Empty:
                if self.finished_processing.stop_check():
                    break
                continue
            self.for_write_queue.task_done()
            if self.stopped.stop_check():
                continue
            try:
                overwrite = True
                file = self.files[file_name]
                output_path = file.construct_output_path(
                    self.output_directory, self.backup_output_directory
                )

                export_textgrid(file, output_path, data, self.frame_shift, overwrite)
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                self.textgrid_errors[file_name] = "\n".join(
                    traceback.format_exception(exc_type, exc_value, exc_traceback)
                )


class ExportPreparationProcessWorker(mp.Process):
    """
    Multiprocessing worker for preparing CTMs for export

    Parameters
    ----------
    to_export_queue: :class:`~multiprocessing.Queue`
        Input queue of combined CTMs
    for_write_queue: :class:`~multiprocessing.Queue`
        Export queue of files to export
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for processing
    finished_combining: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Input signal that all CTMs have been combined
    files: Dict[str, File]
        Files in corpus
    """

    def __init__(
        self,
        to_export_queue: mp.Queue,
        for_write_queue: mp.Queue,
        stopped: Stopped,
        finished_combining: Stopped,
        files: Dict[str, File],
    ):
        mp.Process.__init__(self)
        self.to_export_queue = to_export_queue
        self.for_write_queue = for_write_queue
        self.stopped = stopped
        self.finished_combining = finished_combining

        self.files = files

    def run(self) -> None:
        """Run the export preparation worker"""
        export_data = {}
        try:
            while True:
                try:
                    file_name, data = self.to_export_queue.get(timeout=queue_polling_timeout)
                except Empty:
                    if self.finished_combining.stop_check():
                        break
                    continue
                self.to_export_queue.task_done()
                if self.stopped.stop_check():
                    continue
                file = self.files[file_name]
                if len(file.speaker_ordering) > 1:
                    if file_name not in export_data:
                        export_data[file_name] = data
                    else:
                        export_data[file_name].update(data)
                    if len(export_data[file_name]) == len(file.speaker_ordering):
                        data = export_data.pop(file_name)
                        self.for_write_queue.put((file_name, data))
                else:
                    self.for_write_queue.put((file_name, data))

            for k, v in export_data.items():
                self.for_write_queue.put((k, v))
        except Exception:
            self.stopped.stop()
            raise


def ctms_to_textgrids_mp(aligner: AlignerType):
    """
    Multiprocessing function for exporting alignment CTM information as TextGrids

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    export_begin = time.time()
    manager = mp.Manager()
    textgrid_errors = manager.dict()
    error_catching = manager.dict()
    stopped = Stopped()
    backup_output_directory = None
    if not aligner.align_config.overwrite:
        backup_output_directory = os.path.join(aligner.align_directory, "textgrids")
        os.makedirs(backup_output_directory, exist_ok=True)

    aligner.logger.debug("Beginning to process ctm files...")
    ctm_begin_time = time.time()
    word_procs = []
    phone_procs = []
    combine_procs = []
    finished_signals = [Stopped() for _ in range(aligner.corpus.num_jobs)]
    finished_processing = Stopped()
    to_process_queue = [mp.JoinableQueue() for _ in range(aligner.corpus.num_jobs)]
    to_export_queue = mp.JoinableQueue()
    for_write_queue = mp.JoinableQueue()
    finished_combining = Stopped()
    for j in aligner.corpus.jobs:
        if aligner.align_config.cleanup_textgrids:
            word_p = CleanupWordCtmProcessWorker(
                j.name,
                to_process_queue[j.name],
                stopped,
                error_catching,
                j.cleanup_word_ctm_arguments(aligner),
            )
        else:
            word_p = NoCleanupWordCtmProcessWorker(
                j.name,
                to_process_queue[j.name],
                stopped,
                error_catching,
                j.no_cleanup_word_ctm_arguments(aligner),
            )

        word_procs.append(word_p)
        word_p.start()

        phone_p = PhoneCtmProcessWorker(
            j.name,
            to_process_queue[j.name],
            stopped,
            error_catching,
            j.phone_ctm_arguments(aligner),
        )
        phone_p.start()
        phone_procs.append(phone_p)

        combine_p = CombineProcessWorker(
            j.name,
            to_process_queue[j.name],
            to_export_queue,
            stopped,
            finished_signals[j.name],
            error_catching,
            j.combine_ctm_arguments(aligner),
        )
        combine_p.start()
        combine_procs.append(combine_p)
    preparation_proc = ExportPreparationProcessWorker(
        to_export_queue, for_write_queue, stopped, finished_combining, aligner.corpus.files
    )
    preparation_proc.start()

    export_procs = []
    for j in aligner.corpus.jobs:
        export_proc = ExportTextGridProcessWorker(
            for_write_queue,
            stopped,
            finished_processing,
            textgrid_errors,
            j.export_textgrid_arguments(aligner),
        )
        export_proc.start()
        export_procs.append(export_proc)

    aligner.logger.debug("Waiting for processes to finish...")
    for i in range(aligner.corpus.num_jobs):
        word_procs[i].join()
        phone_procs[i].join()
        finished_signals[i].stop()

    aligner.logger.debug(f"Ctm parsers took {time.time() - ctm_begin_time} seconds")

    aligner.logger.debug("Waiting for processes to finish...")
    for i in range(aligner.corpus.num_jobs):
        to_process_queue[i].join()
        combine_procs[i].join()
    finished_combining.stop()

    to_export_queue.join()
    preparation_proc.join()

    aligner.logger.debug(f"Combiners took {time.time() - ctm_begin_time} seconds")
    aligner.logger.debug("Beginning export...")

    aligner.logger.debug(f"Adding jobs for export took {time.time() - export_begin}")
    aligner.logger.debug("Waiting for export processes to join...")

    for_write_queue.join()
    finished_processing.stop()
    for i in range(aligner.corpus.num_jobs):
        export_procs[i].join()
    for_write_queue.join()
    aligner.logger.debug(f"Export took {time.time() - export_begin} seconds")

    if error_catching:
        aligner.logger.error("Error was encountered in processing CTMs")
        for key, error in error_catching.items():
            aligner.logger.error(f"{key}:\n\n{error}")
        raise AlignmentExportError(error_catching)

    if textgrid_errors:
        aligner.logger.warning(
            f"There were {len(textgrid_errors)} errors encountered in generating TextGrids. "
            f"Check the output_errors.txt file in {os.path.join(aligner.textgrid_output)} "
            f"for more details"
        )
    output_textgrid_writing_errors(aligner.textgrid_output, textgrid_errors)


def convert_ali_to_textgrids(aligner: AlignerType) -> None:
    """
    Multiprocessing function that aligns based on the current model

    See:

    - http://kaldi-asr.org/doc/linear-to-nbest_8cc.html
    - http://kaldi-asr.org/doc/lattice-align-words_8cc.html
    - http://kaldi-asr.org/doc/lattice-to-phone-lattice_8cc.html
    - http://kaldi-asr.org/doc/nbest-to-ctm_8cc.html

    for more details
    on the Kaldi binaries this function calls.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/get_train_ctm.sh
    for the bash script that this function was based on.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    log_directory = aligner.working_log_directory
    os.makedirs(aligner.textgrid_output, exist_ok=True)
    jobs = [x.ali_to_word_ctm_arguments(aligner) for x in aligner.corpus.jobs]  # Word CTM jobs
    jobs += [x.ali_to_phone_ctm_arguments(aligner) for x in aligner.corpus.jobs]  # Phone CTM jobs
    aligner.logger.info("Generating CTMs from alignment...")
    if aligner.use_mp:
        run_mp(ali_to_ctm_func, jobs, log_directory)
    else:
        run_non_mp(ali_to_ctm_func, jobs, log_directory)
    aligner.logger.info("Finished generating CTMs!")

    aligner.logger.info("Exporting TextGrids from CTMs...")
    if aligner.use_mp:
        ctms_to_textgrids_mp(aligner)
    else:
        ctms_to_textgrids_non_mp(aligner)
    aligner.logger.info("Finished exporting TextGrids!")


def tree_stats_func(
    log_path: str,
    dictionaries: List[str],
    ci_phones: str,
    model_path: str,
    feature_strings: Dict[str, str],
    ali_paths: Dict[str, str],
    treeacc_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function for calculating tree stats for training

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    ci_phones: str
        Colon-separated list of context-independent phones
    model_path: str
        Path to the acoustic model file
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    treeacc_paths: Dict[str, str]
        Dictionary of accumulated tree stats files per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            ali_path = ali_paths[dict_name]
            treeacc_path = treeacc_paths[dict_name]
            subprocess.call(
                [
                    thirdparty_binary("acc-tree-stats"),
                    f"--ci-phones={ci_phones}",
                    model_path,
                    feature_string,
                    f"ark:{ali_path}",
                    treeacc_path,
                ],
                stderr=log_file,
            )


def tree_stats(aligner: AlignerType) -> None:
    """
    Multiprocessing function that computes stats for decision tree training

    See http://kaldi-asr.org/doc/acc-tree-stats_8cc.html for more details
    on the Kaldi binary this runs.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """

    jobs = [j.tree_stats_arguments(aligner) for j in aligner.corpus.jobs]

    if aligner.use_mp:
        run_mp(tree_stats_func, jobs, aligner.working_log_directory)
    else:
        run_non_mp(tree_stats_func, jobs, aligner.working_log_directory)

    tree_accs = []
    for x in jobs:
        tree_accs.extend(x.treeacc_paths.values())
    log_path = os.path.join(aligner.working_log_directory, "sum_tree_acc.log")
    with open(log_path, "w", encoding="utf8") as log_file:
        subprocess.call(
            [
                thirdparty_binary("sum-tree-stats"),
                os.path.join(aligner.working_directory, "treeacc"),
            ]
            + tree_accs,
            stderr=log_file,
        )
    if not aligner.debug:
        for f in tree_accs:
            os.remove(f)


def convert_alignments_func(
    log_path: str,
    dictionaries: List[str],
    model_path: str,
    tree_path: str,
    align_model_path: str,
    ali_paths: Dict[str, str],
    new_ali_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function for converting alignments from a previous trainer

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    model_path: str
        Path to the acoustic model file
    tree_path: str
        Path to the acoustic model tree file
    align_model_path: str
        Path to the alignment acoustic model file
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    new_ali_paths: Dict[str, str]
        Dictionary of new alignment archives per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            ali_path = ali_paths[dict_name]
            new_ali_path = new_ali_paths[dict_name]
            subprocess.call(
                [
                    thirdparty_binary("convert-ali"),
                    align_model_path,
                    model_path,
                    tree_path,
                    f"ark:{ali_path}",
                    f"ark:{new_ali_path}",
                ],
                stderr=log_file,
            )


def convert_alignments(aligner: AlignerType) -> None:
    """
    Multiprocessing function that converts alignments from previous training

    See http://kaldi-asr.org/doc/convert-ali_8cc.html for more details
    on the Kaldi binary this runs.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """

    jobs = [x.convert_alignment_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(convert_alignments_func, jobs, aligner.working_log_directory)
    else:
        run_non_mp(convert_alignments_func, jobs, aligner.working_log_directory)


def calc_fmllr_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ali_paths: Dict[str, str],
    ali_model_path: str,
    model_path: str,
    spk2utt_paths: Dict[str, str],
    trans_paths: Dict[str, str],
    fmllr_options: ConfigDict,
) -> None:
    """
    Multiprocessing function for calculating fMLLR transforms

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    ali_model_path: str
        Path to the alignment acoustic model file
    model_path: str
        Path to the acoustic model file
    spk2utt_paths: Dict[str, str]
        Dictionary of spk2utt scps per dictionary name
    trans_paths: Dict[str, str]
        Dictionary of fMLLR transform archives per dictionary name
    fmllr_options: ConfigDict
        Options for fMLLR estimation
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        log_file.writelines(f"{k}: {v}\n" for k, v in os.environ.items())
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            ali_path = ali_paths[dict_name]
            spk2utt_path = spk2utt_paths[dict_name]
            trans_path = trans_paths[dict_name]
            post_proc = subprocess.Popen(
                [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )

            weight_proc = subprocess.Popen(
                [
                    thirdparty_binary("weight-silence-post"),
                    "0.0",
                    fmllr_options["silence_csl"],
                    ali_model_path,
                    "ark:-",
                    "ark:-",
                ],
                stderr=log_file,
                stdin=post_proc.stdout,
                stdout=subprocess.PIPE,
                env=os.environ,
            )

            if ali_model_path != model_path:
                post_gpost_proc = subprocess.Popen(
                    [
                        thirdparty_binary("gmm-post-to-gpost"),
                        ali_model_path,
                        feature_string,
                        "ark:-",
                        "ark:-",
                    ],
                    stderr=log_file,
                    stdin=weight_proc.stdout,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                est_proc = subprocess.Popen(
                    [
                        thirdparty_binary("gmm-est-fmllr-gpost"),
                        "--verbose=4",
                        f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
                        f"--spk2utt=ark:{spk2utt_path}",
                        model_path,
                        feature_string,
                        "ark,s,cs:-",
                        f"ark:{trans_path}",
                    ],
                    stderr=log_file,
                    stdin=post_gpost_proc.stdout,
                    env=os.environ,
                )
                est_proc.communicate()

            else:

                if os.path.exists(trans_path):
                    cmp_trans_path = trans_paths[dict_name] + ".tmp"
                    est_proc = subprocess.Popen(
                        [
                            thirdparty_binary("gmm-est-fmllr"),
                            "--verbose=4",
                            f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
                            f"--spk2utt=ark:{spk2utt_path}",
                            model_path,
                            feature_string,
                            "ark:-",
                            "ark:-",
                        ],
                        stderr=log_file,
                        stdin=weight_proc.stdout,
                        stdout=subprocess.PIPE,
                        env=os.environ,
                    )
                    comp_proc = subprocess.Popen(
                        [
                            thirdparty_binary("compose-transforms"),
                            "--b-is-affine=true",
                            "ark:-",
                            f"ark:{trans_path}",
                            f"ark:{cmp_trans_path}",
                        ],
                        stderr=log_file,
                        stdin=est_proc.stdout,
                        env=os.environ,
                    )
                    comp_proc.communicate()

                    os.remove(trans_path)
                    os.rename(cmp_trans_path, trans_path)
                else:
                    est_proc = subprocess.Popen(
                        [
                            thirdparty_binary("gmm-est-fmllr"),
                            "--verbose=4",
                            f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
                            f"--spk2utt=ark:{spk2utt_path}",
                            model_path,
                            feature_string,
                            "ark,s,cs:-",
                            f"ark:{trans_path}",
                        ],
                        stderr=log_file,
                        stdin=weight_proc.stdout,
                        env=os.environ,
                    )
                    est_proc.communicate()


def calc_fmllr(aligner: AlignerType) -> None:
    """
    Multiprocessing function that computes speaker adaptation (fMLLR)

    See:

    - http://kaldi-asr.org/doc/gmm-est-fmllr_8cc.html
    - http://kaldi-asr.org/doc/ali-to-post_8cc.html
    - http://kaldi-asr.org/doc/weight-silence-post_8cc.html
    - http://kaldi-asr.org/doc/compose-transforms_8cc.html
    - http://kaldi-asr.org/doc/transform-feats_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/align_fmllr.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    begin = time.time()
    log_directory = aligner.working_log_directory

    jobs = [x.calc_fmllr_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(calc_fmllr_func, jobs, log_directory)
    else:
        run_non_mp(calc_fmllr_func, jobs, log_directory)
    aligner.speaker_independent = False
    aligner.logger.debug(f"Fmllr calculation took {time.time() - begin}")


def acc_stats_two_feats_func(
    log_path: str,
    dictionaries: List[str],
    ali_paths: Dict[str, str],
    acc_paths: Dict[str, str],
    model_path: str,
    feature_strings: Dict[str, str],
    si_feature_strings: Dict[str, str],
) -> None:
    """
    Multiprocessing function for accumulating stats across speaker-independent and
    speaker-adapted features

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    acc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    model_path: str
        Path to the acoustic model file
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    si_feature_strings: Dict[str, str]
        Dictionary of speaker-independent feature strings per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            ali_path = ali_paths[dict_name]
            acc_path = acc_paths[dict_name]
            feature_string = feature_strings[dict_name]
            si_feature_string = si_feature_strings[dict_name]
            ali_to_post_proc = subprocess.Popen(
                [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            acc_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-acc-stats-twofeats"),
                    model_path,
                    feature_string,
                    si_feature_string,
                    "ark,s,cs:-",
                    acc_path,
                ],
                stderr=log_file,
                stdin=ali_to_post_proc.stdout,
                env=os.environ,
            )
            acc_proc.communicate()


def create_align_model(aligner: AlignerType) -> None:
    """
    Create alignment model for speaker-adapted training that will use speaker-independent
    features in later aligning

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
        Aligner
    """
    aligner.logger.info("Creating alignment model for speaker-independent features...")
    begin = time.time()
    log_directory = aligner.working_log_directory

    model_path = os.path.join(aligner.working_directory, "final.mdl")
    align_model_path = os.path.join(aligner.working_directory, "final.alimdl")
    arguments = [x.acc_stats_two_feats_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(acc_stats_two_feats_func, arguments, log_directory)
    else:
        run_non_mp(acc_stats_two_feats_func, arguments, log_directory)

    log_path = os.path.join(aligner.working_log_directory, "align_model_est.log")
    with open(log_path, "w", encoding="utf8") as log_file:

        acc_files = []
        for x in arguments:
            acc_files.extend(x.acc_paths.values())
        sum_proc = subprocess.Popen(
            [thirdparty_binary("gmm-sum-accs"), "-"] + acc_files,
            stderr=log_file,
            stdout=subprocess.PIPE,
            env=os.environ,
        )
        est_proc = subprocess.Popen(
            [
                thirdparty_binary("gmm-est"),
                "--remove-low-count-gaussians=false",
                f"--power={aligner.power}",
                model_path,
                "-",
                align_model_path,
            ],
            stdin=sum_proc.stdout,
            stderr=log_file,
            env=os.environ,
        )
        est_proc.communicate()
        if not aligner.debug:
            for f in acc_files:
                os.remove(f)

    aligner.logger.debug(f"Alignment model creation took {time.time() - begin}")


def lda_acc_stats_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ali_paths: Dict[str, str],
    model_path: str,
    lda_options: ConfigDict,
    acc_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function to accumulate LDA stats

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    model_path: str
        Path to the acoustic model file
    lda_options: ConfigDict
        Options for LDA
    acc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            ali_path = ali_paths[dict_name]
            feature_string = feature_strings[dict_name]
            acc_path = acc_paths[dict_name]
            ali_to_post_proc = subprocess.Popen(
                [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            weight_silence_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("weight-silence-post"),
                    f"{lda_options['boost_silence']}",
                    lda_options["silence_csl"],
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=ali_to_post_proc.stdout,
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            acc_lda_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("acc-lda"),
                    f"--rand-prune={lda_options['random_prune']}",
                    model_path,
                    feature_string,
                    "ark,s,cs:-",
                    acc_path,
                ],
                stdin=weight_silence_post_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            acc_lda_post_proc.communicate()


def lda_acc_stats(aligner: LdaTrainer) -> None:
    """
    Multiprocessing function that accumulates LDA statistics

    See:

    - http://kaldi-asr.org/doc/ali-to-post_8cc.html
    - http://kaldi-asr.org/doc/weight-silence-post_8cc.html
    - http://kaldi-asr.org/doc/acc-lda_8cc.html
    - http://kaldi-asr.org/doc/est-lda_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_lda_mllt.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.LdaTrainer`
        Trainer
    """
    arguments = [x.lda_acc_stats_arguments(aligner) for x in aligner.corpus.jobs]

    if aligner.use_mp:
        run_mp(lda_acc_stats_func, arguments, aligner.working_log_directory)
    else:
        run_non_mp(lda_acc_stats_func, arguments, aligner.working_log_directory)

    log_path = os.path.join(aligner.working_log_directory, "lda_est.log")
    acc_list = []
    for x in arguments:
        acc_list.extend(x.acc_paths.values())
    with open(log_path, "w", encoding="utf8") as log_file:
        est_lda_proc = subprocess.Popen(
            [
                thirdparty_binary("est-lda"),
                f"--write-full-matrix={os.path.join(aligner.working_directory, 'full.mat')}",
                f"--dim={aligner.lda_dimension}",
                os.path.join(aligner.working_directory, "lda.mat"),
            ]
            + acc_list,
            stderr=log_file,
            env=os.environ,
        )
        est_lda_proc.communicate()


def calc_lda_mllt_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ali_paths: Dict[str, str],
    model_path: str,
    lda_options: ConfigDict,
    macc_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function for estimating LDA with MLLT

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    model_path: str
        Path to the acoustic model file
    lda_options: ConfigDict
        Options for LDA
    macc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    """
    # Estimating MLLT
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            ali_path = ali_paths[dict_name]
            feature_string = feature_strings[dict_name]
            macc_path = macc_paths[dict_name]
            post_proc = subprocess.Popen(
                [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )

            weight_proc = subprocess.Popen(
                [
                    thirdparty_binary("weight-silence-post"),
                    "0.0",
                    lda_options["silence_csl"],
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=post_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            acc_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-acc-mllt"),
                    f"--rand-prune={lda_options['random_prune']}",
                    model_path,
                    feature_string,
                    "ark,s,cs:-",
                    macc_path,
                ],
                stdin=weight_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            acc_proc.communicate()


def calc_lda_mllt(aligner: LdaTrainer) -> None:
    """
    Multiprocessing function that calculates LDA+MLLT transformations

    See:

    - http://kaldi-asr.org/doc/ali-to-post_8cc.html
    - http://kaldi-asr.org/doc/weight-silence-post_8cc.html
    - http://kaldi-asr.org/doc/gmm-acc-mllt_8cc.html
    - http://kaldi-asr.org/doc/est-mllt_8cc.html
    - http://kaldi-asr.org/doc/gmm-transform-means_8cc.html
    - http://kaldi-asr.org/doc/compose-transforms_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_lda_mllt.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.trainer.LdaTrainer`
        Trainer
    """
    jobs = [x.calc_lda_mllt_arguments(aligner) for x in aligner.corpus.jobs]

    if aligner.use_mp:
        run_mp(calc_lda_mllt_func, jobs, aligner.working_log_directory)
    else:
        run_non_mp(calc_lda_mllt_func, jobs, aligner.working_log_directory)

    log_path = os.path.join(
        aligner.working_log_directory, f"transform_means.{aligner.iteration}.log"
    )
    previous_mat_path = os.path.join(aligner.working_directory, "lda.mat")
    new_mat_path = os.path.join(aligner.working_directory, "lda_new.mat")
    composed_path = os.path.join(aligner.working_directory, "lda_composed.mat")
    with open(log_path, "a", encoding="utf8") as log_file:
        macc_list = []
        for x in jobs:
            macc_list.extend(x.macc_paths.values())
        subprocess.call(
            [thirdparty_binary("est-mllt"), new_mat_path] + macc_list,
            stderr=log_file,
            env=os.environ,
        )
        subprocess.call(
            [
                thirdparty_binary("gmm-transform-means"),
                new_mat_path,
                aligner.current_model_path,
                aligner.current_model_path,
            ],
            stderr=log_file,
            env=os.environ,
        )

        if os.path.exists(previous_mat_path):
            subprocess.call(
                [
                    thirdparty_binary("compose-transforms"),
                    new_mat_path,
                    previous_mat_path,
                    composed_path,
                ],
                stderr=log_file,
                env=os.environ,
            )
            os.remove(previous_mat_path)
            os.rename(composed_path, previous_mat_path)
        else:
            os.rename(new_mat_path, previous_mat_path)


def map_acc_stats_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    model_path: str,
    ali_paths: Dict[str, str],
    acc_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function for accumulating mapped stats for adapting acoustic models to new
    domains

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    model_path: str
        Path to the acoustic model file
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    acc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            acc_path = acc_paths[dict_name]
            ali_path = ali_paths[dict_name]
            acc_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-acc-stats-ali"),
                    model_path,
                    feature_string,
                    f"ark,s,cs:{ali_path}",
                    acc_path,
                ],
                stderr=log_file,
                env=os.environ,
            )
            acc_proc.communicate()


def train_map(aligner: AdaptingAligner) -> None:
    """
    Trains an adapted acoustic model through mapping model states and update those with
    enough data

    Source: https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_map.sh

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.aligner.AdaptingAligner`
        Adapting aligner
    """
    begin = time.time()
    initial_mdl_path = os.path.join(aligner.working_directory, "0.mdl")
    final_mdl_path = os.path.join(aligner.working_directory, "final.mdl")
    log_directory = aligner.working_log_directory
    os.makedirs(log_directory, exist_ok=True)

    jobs = [x.map_acc_stats_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.use_mp:
        run_mp(map_acc_stats_func, jobs, log_directory)
    else:
        run_non_mp(map_acc_stats_func, jobs, log_directory)
    log_path = os.path.join(aligner.working_log_directory, "map_model_est.log")
    occs_path = os.path.join(aligner.working_directory, "final.occs")
    with open(log_path, "w", encoding="utf8") as log_file:
        acc_files = []
        for j in jobs:
            acc_files.extend(j.acc_paths.values())
        sum_proc = subprocess.Popen(
            [thirdparty_binary("gmm-sum-accs"), "-"] + acc_files,
            stderr=log_file,
            stdout=subprocess.PIPE,
            env=os.environ,
        )
        ismooth_proc = subprocess.Popen(
            [
                thirdparty_binary("gmm-ismooth-stats"),
                "--smooth-from-model",
                f"--tau={aligner.mapping_tau}",
                initial_mdl_path,
                "-",
                "-",
            ],
            stderr=log_file,
            stdin=sum_proc.stdout,
            stdout=subprocess.PIPE,
            env=os.environ,
        )
        est_proc = subprocess.Popen(
            [
                thirdparty_binary("gmm-est"),
                "--update-flags=m",
                f"--write-occs={occs_path}",
                "--remove-low-count-gaussians=false",
                initial_mdl_path,
                "-",
                final_mdl_path,
            ],
            stdin=ismooth_proc.stdout,
            stderr=log_file,
            env=os.environ,
        )
        est_proc.communicate()

    initial_alimdl_path = os.path.join(aligner.working_directory, "0.alimdl")
    final_alimdl_path = os.path.join(aligner.working_directory, "0.alimdl")
    if os.path.exists(initial_alimdl_path):
        aligner.speaker_independent = True
        jobs = [x.map_acc_stats_arguments(aligner) for x in aligner.corpus.jobs]
        if aligner.use_mp:
            run_mp(map_acc_stats_func, jobs, log_directory)
        else:
            run_non_mp(map_acc_stats_func, jobs, log_directory)

        log_path = os.path.join(aligner.working_log_directory, "map_model_est.log")
        with open(log_path, "w", encoding="utf8") as log_file:
            acc_files = []
        for j in jobs:
            acc_files.extend(j.acc_paths)
            sum_proc = subprocess.Popen(
                [thirdparty_binary("gmm-sum-accs"), "-"] + acc_files,
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            ismooth_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-ismooth-stats"),
                    "--smooth-from-model",
                    f"--tau={aligner.mapping_tau}",
                    initial_alimdl_path,
                    "-",
                    "-",
                ],
                stderr=log_file,
                stdin=sum_proc.stdout,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            est_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-est"),
                    "--update-flags=m",
                    "--remove-low-count-gaussians=false",
                    initial_alimdl_path,
                    "-",
                    final_alimdl_path,
                ],
                stdin=ismooth_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            est_proc.communicate()

    aligner.logger.debug(f"Mapping models took {time.time() - begin}")


def test_utterances_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    words_paths: Dict[str, str],
    graphs_paths: Dict[str, str],
    text_int_paths: Dict[str, str],
    edits_paths: Dict[str, str],
    out_int_paths: Dict[str, str],
    model_path: str,
):
    """
    Multiprocessing function to test utterance transcriptions

    Parameters
    ----------
    log_path: str
        Log path
    dictionaries: List[str]
        List of dictionaries
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    words_paths: Dict[str, str]
        Dictionary of word mapping files per dictionary name
    graphs_paths: Dict[str, str]
        Dictionary of utterance FST graph archives per dictionary name
    text_int_paths: Dict[str, str]
        Dictionary of text.int files per dictionary name
    edits_paths: Dict[str, str]
        Dictionary of paths to save transcription differences per dictionary name
    out_int_paths: Dict[str, str]
        Dictionary of output .int files per dictionary name
    model_path: str
        Acoustic model path
    """
    acoustic_scale = 0.1
    beam = 15.0
    lattice_beam = 8.0
    max_active = 750
    with open(log_path, "w") as log_file:
        for dict_name in dictionaries:
            words_path = words_paths[dict_name]
            graphs_path = graphs_paths[dict_name]
            feature_string = feature_strings[dict_name]
            edits_path = edits_paths[dict_name]
            text_int_path = text_int_paths[dict_name]
            out_int_path = out_int_paths[dict_name]
            latgen_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-latgen-faster"),
                    f"--acoustic-scale={acoustic_scale}",
                    f"--beam={beam}",
                    f"--max-active={max_active}",
                    f"--lattice-beam={lattice_beam}",
                    f"--word-symbol-table={words_path}",
                    model_path,
                    "ark:" + graphs_path,
                    feature_string,
                    "ark:-",
                ],
                stderr=log_file,
                stdout=subprocess.PIPE,
            )

            oracle_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-oracle"),
                    "ark:-",
                    f"ark,t:{text_int_path}",
                    f"ark,t:{out_int_path}",
                    f"ark,t:{edits_path}",
                ],
                stderr=log_file,
                stdin=latgen_proc.stdout,
            )
            oracle_proc.communicate()


def compile_utterance_train_graphs_func(
    log_path: str,
    dictionaries: List[str],
    disambig_int_paths: Dict[str, str],
    disambig_L_fst_paths: Dict[str, str],
    fst_paths: Dict[str, str],
    graphs_paths: Dict[str, str],
    model_path: str,
    tree_path: str,
):
    """
    Multiprocessing function to compile utterance FSTs

    Parameters
    ----------
    log_path: str
        Log path
    dictionaries: List[str]
        List of dictionaries
    disambig_int_paths: Dict[str, str]
        Dictionary of disambiguation symbol int files per dictionary name
    disambig_L_fst_paths: Dict[str, str]
        Dictionary of disambiguation lexicon FSTs per dictionary name
    fst_paths: Dict[str, str]
        Dictionary of pregenerated utterance FST scp files per dictionary name
    graphs_paths: Dict[str, str]
        Dictionary of utterance FST graph archives per dictionary name
    model_path: str
        Acoustic model path
    tree_path: str
        Acoustic model's tree path
    """
    with open(log_path, "w") as log_file:
        for dict_name in dictionaries:
            disambig_int_path = disambig_int_paths[dict_name]
            disambig_L_fst_path = disambig_L_fst_paths[dict_name]
            fst_path = fst_paths[dict_name]
            graphs_path = graphs_paths[dict_name]
            proc = subprocess.Popen(
                [
                    thirdparty_binary("compile-train-graphs-fsts"),
                    "--transition-scale=1.0",
                    "--self-loop-scale=0.1",
                    f"--read-disambig-syms={disambig_int_path}",
                    tree_path,
                    model_path,
                    disambig_L_fst_path,
                    f"ark:{fst_path}",
                    f"ark:{graphs_path}",
                ],
                stderr=log_file,
            )

            proc.communicate()



"""Class definitions for Monophone trainer"""
from __future__ import annotations

import os
import re
import subprocess
import time
from typing import TYPE_CHECKING, Optional

from ..exceptions import KaldiProcessingError
from ..multiprocessing import compile_train_graphs, mono_align_equal
from ..utils import log_kaldi_errors, parse_logs, thirdparty_binary
from .base import BaseTrainer

if TYPE_CHECKING:
    from ..config import FeatureConfig
    from ..corpus import Corpus
    from ..dictionary import DictionaryType


__all__ = ["MonophoneTrainer"]


class MonophoneTrainer(BaseTrainer):
    """
    Configuration class for monophone training


    Attributes
    ----------
    initial_gaussians : int
        Number of gaussians to begin training
    """

    def __init__(self, default_feature_config: FeatureConfig):
        super(MonophoneTrainer, self).__init__(default_feature_config)
        self.initial_gaussians = 135
        self.compute_calculated_properties()

    def compute_calculated_properties(self) -> None:
        """Generate realignment iterations and initial gaussians based on configuration"""
        for i in range(1, self.num_iterations):
            if i <= int(self.num_iterations / 4):
                self.realignment_iterations.append(i)
            elif i <= int(self.num_iterations * 2 / 4):
                if i - self.realignment_iterations[-1] > 1:
                    self.realignment_iterations.append(i)
            else:
                if i - self.realignment_iterations[-1] > 2:
                    self.realignment_iterations.append(i)

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "mono"

    @property
    def phone_type(self) -> str:
        """Phone type"""
        return "monophone"

    def get_num_gauss(self) -> int:
        """
        Get the number of gaussians for a monophone model

        Returns
        -------
        int
            Initial number of gaussians
        """
        with open(os.devnull, "w") as devnull:
            proc = subprocess.Popen(
                [thirdparty_binary("gmm-info"), "--print-args=false", self.current_model_path],
                stderr=devnull,
                stdout=subprocess.PIPE,
            )
            stdout, stderr = proc.communicate()
            num = stdout.decode("utf8")
            matches = re.search(r"gaussians (\d+)", num)
            num = int(matches.groups()[0])
        return num

    def init_training(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[BaseTrainer] = None,
    ) -> None:
        """
        Initialize monophone training

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: TrainerType, optional
            Previous trainer to initialize from

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        self._setup_for_init(identifier, temporary_directory, corpus, dictionary, previous_trainer)
        done_path = os.path.join(self.train_directory, "done")
        dirty_path = os.path.join(self.train_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info(f"{self.identifier} training already done, skipping initialization.")
            return
        begin = time.time()
        self.iteration = 0
        tree_path = os.path.join(self.train_directory, "tree")

        try:
            feat_dim = corpus.get_feat_dim()

            feature_string = corpus.jobs[0].construct_base_feature_string(corpus)
            shared_phones_path = os.path.join(dictionary.phones_dir, "sets.int")
            init_log_path = os.path.join(self.log_directory, "init.log")
            temp_feats_path = os.path.join(self.train_directory, "temp_feats")
            with open(init_log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("subset-feats"),
                        "--n=10",
                        feature_string,
                        f"ark:{temp_feats_path}",
                    ],
                    stderr=log_file,
                )
                subprocess.call(
                    [
                        thirdparty_binary("gmm-init-mono"),
                        f"--shared-phones={shared_phones_path}",
                        f"--train-feats=ark:{temp_feats_path}",
                        os.path.join(dictionary.output_directory, "topo"),
                        str(feat_dim),
                        self.current_model_path,
                        tree_path,
                    ],
                    stderr=log_file,
                )
            if os.path.exists(self.current_model_path):
                os.remove(init_log_path)
            os.remove(temp_feats_path)
            num_gauss = self.get_num_gauss()
            self.initial_gaussians = num_gauss
            self.current_gaussians = num_gauss
            compile_train_graphs(self)
            mono_align_equal(self)
            self.iteration = 1
            parse_logs(self.log_directory)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        self.logger.info("Initialization complete!")
        self.logger.debug(f"Initialization took {time.time() - begin} seconds")



"""Class for generating pronunciations from G2P models"""
from __future__ import annotations

import functools
import logging
import multiprocessing as mp
import os
import queue
import sys
import time
import traceback
from typing import TYPE_CHECKING, Any, Collection, Dict, List, Optional, Set, Tuple, Union

import tqdm

from ..config import TEMP_DIR
from ..exceptions import G2PError
from ..multiprocessing import Counter, Stopped

try:
    import pynini
    from pynini import Fst, TokenType
    from pynini.lib import rewrite

    G2P_DISABLED = False
except ImportError:
    pynini = None
    TokenType = None
    Fst = None
    rewrite = None
    G2P_DISABLED = True

if TYPE_CHECKING:
    SpeakerCharacterType = Union[str, int]
    from ..models import G2PModel


__all__ = ["Rewriter", "RewriterWorker", "PyniniDictionaryGenerator"]


class Rewriter:
    """Helper object for rewriting."""

    def __init__(
        self, fst: Fst, input_token_type: TokenType, output_token_type: TokenType, nshortest=1
    ):
        self.rewrite = functools.partial(
            rewrite.top_rewrites,
            nshortest=nshortest,
            rule=fst,
            input_token_type=input_token_type,
            output_token_type=output_token_type,
        )

    def __call__(self, i: str) -> str:  # pragma: no cover
        """Call the rewrite function"""
        try:
            return self.rewrite(i)
        except rewrite.Error:
            return "<composition failure>"


class RewriterWorker(mp.Process):
    """
    Rewriter process
    """

    def __init__(
        self,
        job_q: mp.Queue,
        return_dict: Dict[str, Union[str, Any]],
        rewriter: Rewriter,
        counter: Counter,
        stopped: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_q = job_q
        self.return_dict = return_dict
        self.rewriter = rewriter
        self.counter = counter
        self.stopped = stopped

    def run(self) -> None:
        """Run the rewriting function"""
        while True:
            try:
                word = self.job_q.get(timeout=1)
            except queue.Empty:
                break
            self.job_q.task_done()
            if self.stopped.stop_check():
                continue
            try:
                rep = self.rewriter.rewrite(word)
                if rep != "<composition failure>":
                    self.return_dict[word] = rep
            except rewrite.Error:
                pass
            except Exception:
                self.stopped.stop()
                self.return_dict["MFA_EXCEPTION"] = word, Exception(
                    traceback.format_exception(*sys.exc_info())
                )
            self.counter.increment()
        return


def clean_up_word(word: str, graphemes: Set[str]) -> Tuple[str, List[str]]:
    """
    Clean up word by removing graphemes not in a specified set

    Parameters
    ----------
    word : str
        Input string
    graphemes: Set[str]
        Set of allowable graphemes

    Returns
    -------
    str
        Cleaned up word
    list
        Graphemes excluded
    """
    new_word = []
    missing_graphemes = []
    for c in word:
        if c not in graphemes:
            missing_graphemes.append(c)
        else:
            new_word.append(c)
    return "".join(new_word), missing_graphemes


class PyniniDictionaryGenerator:
    """
    Class for generating pronunciations from a G2P model
    """

    def __init__(
        self,
        g2p_model: G2PModel,
        word_set: Collection[str],
        temp_directory: Optional[str] = None,
        num_jobs: int = 3,
        num_pronunciations: int = 1,
        logger: Optional[logging.Logger] = None,
    ):
        super(PyniniDictionaryGenerator, self).__init__()
        if not temp_directory:
            temp_directory = TEMP_DIR
        temp_directory = os.path.join(temp_directory, "G2P")
        self.model = g2p_model

        self.temp_directory = os.path.join(temp_directory, self.model.name)
        log_dir = os.path.join(self.temp_directory, "logging")
        os.makedirs(log_dir, exist_ok=True)
        self.log_file = os.path.join(log_dir, "g2p.log")
        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger("g2p")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
        self.words = word_set
        self.num_jobs = num_jobs
        self.num_pronunciations = num_pronunciations

    def generate(self) -> Dict[str, List[str]]:
        """
        Generate pronunciations

        Returns
        -------
        Dict
            Mappings of keys to their generated pronunciations
        """
        if self.model.meta["architecture"] == "phonetisaurus":
            raise G2PError(
                "Previously trained Phonetisaurus models from 1.1 and earlier are not currently supported. "
                "Please retrain your model using 2.0+"
            )

        input_token_type = "utf8"
        fst = pynini.Fst.read(self.model.fst_path)

        output_token_type = "utf8"
        if self.model.sym_path is not None and os.path.exists(self.model.sym_path):
            output_token_type = pynini.SymbolTable.read_text(self.model.sym_path)
        rewriter = Rewriter(fst, input_token_type, output_token_type, self.num_pronunciations)

        ind = 0
        num_words = len(self.words)
        words = list(self.words)
        begin = time.time()
        last_value = 0
        missing_graphemes = set()
        print("Generating pronunciations...")
        to_return = {}
        if num_words < 30 or self.num_jobs < 2:
            for word in words:
                w, m = clean_up_word(word, self.model.meta["graphemes"])
                missing_graphemes.update(m)
                if not w:
                    continue
                try:
                    pron = rewriter.rewrite(w)
                except rewrite.Error:
                    continue
                to_return[word] = pron
        else:
            stopped = Stopped()
            job_queue = mp.JoinableQueue(100)
            with tqdm.tqdm(total=num_words) as pbar:
                while True:
                    if ind == num_words:
                        break
                    try:
                        w, m = clean_up_word(words[ind], self.model.meta["graphemes"])
                        missing_graphemes.update(m)
                        if not w:
                            ind += 1
                            continue
                        job_queue.put(w, False)
                    except queue.Full:
                        break
                    ind += 1
                manager = mp.Manager()
                return_dict = manager.dict()
                procs = []
                counter = Counter()
                for _ in range(self.num_jobs):
                    p = RewriterWorker(job_queue, return_dict, rewriter, counter, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    if ind == num_words:
                        break
                    w, m = clean_up_word(words[ind], self.model.meta["graphemes"])
                    missing_graphemes.update(m)
                    if not w:
                        ind += 1
                        continue
                    job_queue.put(w)
                    value = counter.value()
                    pbar.update(value - last_value)
                    last_value = value
                    ind += 1
                job_queue.join()
            for p in procs:
                p.join()
            if "MFA_EXCEPTION" in return_dict:
                element, exc = return_dict["MFA_EXCEPTION"]
                print(element)
                raise exc
            for w in self.words:
                if w in return_dict:
                    to_return[w] = return_dict[w]
        self.logger.debug(f"Processed {num_words} in {time.time() - begin} seconds")
        return to_return

    def output(self, outfile: str) -> None:
        """
        Output pronunciations to text file

        Parameters
        ----------
        outfile: str
            Path to save
        """
        results = self.generate()
        with open(outfile, "w", encoding="utf8") as f:
            for (word, pronunciation) in results.items():
                if not pronunciation:
                    continue
                if isinstance(pronunciation, list):
                    for p in pronunciation:
                        if not p:
                            continue
                        f.write(f"{word}\t{p}\n")
                else:
                    f.write(f"{word}\t{pronunciation}\n")



"""Command line functions for transcribing corpora"""
from __future__ import annotations

import os
import shutil
import time
from typing import TYPE_CHECKING, Optional

from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.config import (
    TEMP_DIR,
    load_basic_transcribe,
    load_command_configuration,
    transcribe_yaml_to_config,
)
from montreal_forced_aligner.corpus import Corpus
from montreal_forced_aligner.dictionary import Dictionary, MultispeakerDictionary
from montreal_forced_aligner.exceptions import ArgumentError
from montreal_forced_aligner.models import AcousticModel, LanguageModel
from montreal_forced_aligner.transcriber import Transcriber
from montreal_forced_aligner.utils import log_config, setup_logger

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["transcribe_corpus", "validate_args", "run_transcribe_corpus"]


def transcribe_corpus(args: Namespace, unknown_args: Optional[list] = None) -> None:
    """
    Run the transcription command

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    from montreal_forced_aligner.utils import get_mfa_version

    command = "transcribe"
    all_begin = time.time()
    if not args.temp_directory:
        temp_dir = TEMP_DIR
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    corpus_name = os.path.basename(args.corpus_directory)
    if corpus_name == "":
        args.corpus_directory = os.path.dirname(args.corpus_directory)
        corpus_name = os.path.basename(args.corpus_directory)
    if args.config_path:
        transcribe_config = transcribe_yaml_to_config(args.config_path)
    else:
        transcribe_config = load_basic_transcribe()
    transcribe_config.use_mp = not args.disable_mp
    transcribe_config.overwrite = args.overwrite
    if unknown_args:
        transcribe_config.update_from_unknown_args(unknown_args)
    data_directory = os.path.join(temp_dir, corpus_name)
    if getattr(args, "clean", False) and os.path.exists(data_directory):
        print("Cleaning old directory!")
        shutil.rmtree(data_directory, ignore_errors=True)
    if getattr(args, "verbose", False):
        log_level = "debug"
    else:
        log_level = "info"
    logger = setup_logger(command, data_directory, console_level=log_level)
    logger.debug("TRANSCRIBE CONFIG:")
    log_config(logger, transcribe_config)
    os.makedirs(data_directory, exist_ok=True)
    model_directory = os.path.join(data_directory, "acoustic_models")
    os.makedirs(args.output_directory, exist_ok=True)
    os.makedirs(model_directory, exist_ok=True)
    conf_path = os.path.join(data_directory, "config.yml")
    conf = load_command_configuration(
        conf_path,
        {
            "dirty": False,
            "begin": time.time(),
            "version": get_mfa_version(),
            "type": "transcribe",
            "corpus_directory": args.corpus_directory,
            "dictionary_path": args.dictionary_path,
            "acoustic_model_path": args.acoustic_model_path,
            "language_model_path": args.language_model_path,
        },
    )

    if (
        conf["dirty"]
        or conf["type"] != command
        or conf["corpus_directory"] != args.corpus_directory
        or conf["version"] != get_mfa_version()
        or conf["dictionary_path"] != args.dictionary_path
        or conf["language_model_path"] != args.language_model_path
        or conf["acoustic_model_path"] != args.acoustic_model_path
    ):
        logger.warning(
            "WARNING: Using old temp directory, this might not be ideal for you, use the --clean flag to ensure no "
            "weird behavior for previous versions of the temporary directory."
        )
        if conf["dirty"]:
            logger.debug("Previous run ended in an error (maybe ctrl-c?)")
        if conf["type"] != command:
            logger.debug(
                f"Previous run was a different subcommand than {command} (was {conf['type']})"
            )
        if conf["corpus_directory"] != args.corpus_directory:
            logger.debug(
                "Previous run used source directory "
                f"path {conf['corpus_directory']} (new run: {args.corpus_directory})"
            )
        if conf["version"] != get_mfa_version():
            logger.debug(
                f"Previous run was on {conf['version']} version (new run: {get_mfa_version()})"
            )
        if conf["dictionary_path"] != args.dictionary_path:
            logger.debug(
                f"Previous run used dictionary path {conf['dictionary_path']} "
                f"(new run: {args.dictionary_path})"
            )
        if conf["acoustic_model_path"] != args.acoustic_model_path:
            logger.debug(
                f"Previous run used acoustic model path {conf['acoustic_model_path']} "
                f"(new run: {args.acoustic_model_path})"
            )
        if conf["language_model_path"] != args.language_model_path:
            logger.debug(
                f"Previous run used language model path {conf['language_model_path']} "
                f"(new run: {args.language_model_path})"
            )
    audio_dir = None
    if args.audio_directory:
        audio_dir = args.audio_directory
    try:
        corpus = Corpus(
            args.corpus_directory,
            data_directory,
            speaker_characters=args.speaker_characters,
            sample_rate=transcribe_config.feature_config.sample_frequency,
            num_jobs=args.num_jobs,
            use_mp=transcribe_config.use_mp,
            audio_directory=audio_dir,
            ignore_speakers=transcribe_config.ignore_speakers,
        )
        acoustic_model = AcousticModel(args.acoustic_model_path, root_directory=model_directory)
        acoustic_model.log_details(logger)
        if args.language_model_path.endswith(".arpa"):
            alternative_name = os.path.splitext(args.language_model_path)[0] + ".zip"
            logger.warning(
                f"Using a plain .arpa model requires generating pruned versions of it to decode in a reasonable "
                f"amount of time.  If you'd like to generate a reusable language model, consider running "
                f"`mfa train_lm {args.language_model_path} {alternative_name}`."
            )
        language_model = LanguageModel(args.language_model_path, root_directory=data_directory)
        if args.dictionary_path.lower().endswith(".yaml"):
            dictionary = MultispeakerDictionary(
                args.dictionary_path,
                data_directory,
                logger=logger,
                punctuation=transcribe_config.punctuation,
                clitic_markers=transcribe_config.clitic_markers,
                compound_markers=transcribe_config.compound_markers,
                multilingual_ipa=acoustic_model.meta["multilingual_ipa"],
            )
        else:
            dictionary = Dictionary(
                args.dictionary_path,
                data_directory,
                logger=logger,
                punctuation=transcribe_config.punctuation,
                clitic_markers=transcribe_config.clitic_markers,
                compound_markers=transcribe_config.compound_markers,
                multilingual_ipa=acoustic_model.meta["multilingual_ipa"],
            )
        acoustic_model.validate(dictionary)
        begin = time.time()
        t = Transcriber(
            corpus,
            dictionary,
            acoustic_model,
            language_model,
            transcribe_config,
            temp_directory=data_directory,
            debug=getattr(args, "debug", False),
            evaluation_mode=args.evaluate,
            logger=logger,
        )
        logger.debug(f"Setup transcriber in {time.time() - begin} seconds")

        begin = time.time()
        t.transcribe()
        logger.debug(f"Performed transcribing in {time.time() - begin} seconds")
        if args.evaluate:
            t.evaluate()
            best_config_path = os.path.join(args.output_directory, "best_transcribe_config.yaml")
            t.transcribe_config.save(best_config_path)
            t.export_transcriptions(args.output_directory)
        else:
            begin = time.time()
            t.export_transcriptions(args.output_directory)
            logger.debug(f"Exported transcriptions in {time.time() - begin} seconds")
        logger.info(f"Done! Everything took {time.time() - all_begin} seconds")
    except Exception:
        conf["dirty"] = True
        raise
    finally:
        handlers = logger.handlers[:]
        for handler in handlers:
            handler.close()
            logger.removeHandler(handler)
        conf.save(conf_path)


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    try:
        args.speaker_characters = int(args.speaker_characters)
    except ValueError:
        pass
    args.output_directory = args.output_directory.rstrip("/").rstrip("\\")
    args.corpus_directory = args.corpus_directory.rstrip("/").rstrip("\\")

    args.dictionary_path = validate_model_arg(args.dictionary_path, "dictionary")
    args.acoustic_model_path = validate_model_arg(args.acoustic_model_path, "acoustic")
    args.language_model_path = validate_model_arg(args.language_model_path, "language_model")

    if not os.path.exists(args.corpus_directory):
        raise ArgumentError(f"Could not find the corpus directory {args.corpus_directory}.")
    if not os.path.isdir(args.corpus_directory):
        raise ArgumentError(
            f"The specified corpus directory ({args.corpus_directory}) is not a directory."
        )

    if args.corpus_directory == args.output_directory:
        raise ArgumentError("Corpus directory and output directory cannot be the same folder.")


def run_transcribe_corpus(args: Namespace, unknown: Optional[list] = None) -> None:
    """
    Wrapper function for running corpus transcription

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    transcribe_corpus(args, unknown)



"""Command line functions for training language models"""
from __future__ import annotations

import os
import shutil
import time
from typing import TYPE_CHECKING, Optional

from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.config import TEMP_DIR, load_basic_train_lm, train_lm_yaml_to_config
from montreal_forced_aligner.corpus import Corpus
from montreal_forced_aligner.dictionary import Dictionary
from montreal_forced_aligner.exceptions import ArgumentError
from montreal_forced_aligner.lm.trainer import LmTrainer
from montreal_forced_aligner.utils import setup_logger

if TYPE_CHECKING:
    from argparse import Namespace

__all__ = ["train_lm", "validate_args", "run_train_lm"]


def train_lm(args: Namespace, unknown_args: Optional[list] = None) -> None:
    """
    Run the language model training

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    command = "train_lm"
    all_begin = time.time()
    if not args.temp_directory:
        temp_dir = TEMP_DIR
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    if args.config_path:
        train_config = train_lm_yaml_to_config(args.config_path)
    else:
        train_config = load_basic_train_lm()
    train_config.use_mp = not args.disable_mp
    if unknown_args:
        train_config.update_from_unknown_args(unknown_args)
    corpus_name = os.path.basename(args.source_path)
    if corpus_name == "":
        args.source_path = os.path.dirname(args.source_path)
        corpus_name = os.path.basename(args.source_path)
    source = args.source_path
    dictionary = None
    if args.source_path.lower().endswith(".arpa"):
        corpus_name = os.path.splitext(corpus_name)[0]
        data_directory = os.path.join(temp_dir, corpus_name)
    else:
        data_directory = os.path.join(temp_dir, corpus_name)
    if getattr(args, "clean", False) and os.path.exists(data_directory):
        print("Cleaning old directory!")
        shutil.rmtree(data_directory, ignore_errors=True)

    if getattr(args, "verbose", False):
        log_level = "debug"
    else:
        log_level = "info"
    logger = setup_logger(command, data_directory, console_level=log_level)
    if not args.source_path.lower().endswith(".arpa"):
        source = Corpus(
            args.source_path,
            data_directory,
            num_jobs=args.num_jobs,
            use_mp=train_config.use_mp,
            parse_text_only_files=True,
            debug=args.debug,
        )
        if args.dictionary_path:
            dictionary = Dictionary(
                args.dictionary_path, data_directory, debug=args.debug, word_set=source.word_set
            )
            dictionary.generate_mappings()
        else:
            dictionary = None
    trainer = LmTrainer(
        source,
        train_config,
        args.output_model_path,
        dictionary=dictionary,
        temp_directory=data_directory,
        supplemental_model_path=args.model_path,
        supplemental_model_weight=args.model_weight,
        debug=args.debug,
        logger=logger,
    )
    begin = time.time()
    trainer.train()
    logger.debug(f"Training took {time.time() - begin} seconds")

    logger.info("All done!")
    logger.debug(f"Done! Everything took {time.time() - all_begin} seconds")
    handlers = logger.handlers[:]
    for handler in handlers:
        handler.close()
        logger.removeHandler(handler)


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    args.source_path = args.source_path.rstrip("/").rstrip("\\")
    if args.dictionary_path:
        args.dictionary_path = validate_model_arg(args.dictionary_path, "dictionary")
    if not args.source_path.endswith(".arpa"):
        if not os.path.exists(args.source_path):
            raise (ArgumentError(f"Could not find the corpus directory {args.source_path}."))
        if not os.path.isdir(args.source_path):
            raise (
                ArgumentError(
                    f"The specified corpus directory ({args.source_path}) is not a directory."
                )
            )
    else:
        if not os.path.exists(args.source_path):
            raise (ArgumentError(f"Could not find the source file {args.source_path}."))
    if args.config_path and not os.path.exists(args.config_path):
        raise (ArgumentError(f"Could not find the config file {args.config_path}."))
    if args.model_path and not os.path.exists(args.model_path):
        raise (ArgumentError(f"Could not find the model file {args.model_path}."))


def run_train_lm(args: Namespace, unknown: Optional[list] = None) -> None:
    """
    Wrapper function for running language model training

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    train_lm(args, unknown)



"""Class definitions for base aligner"""
from __future__ import annotations

import logging
import os
import shutil
import time
from typing import TYPE_CHECKING, Optional

from ..config import TEMP_DIR
from ..exceptions import KaldiProcessingError
from ..multiprocessing import (
    align,
    calc_fmllr,
    compile_information,
    compile_train_graphs,
    convert_ali_to_textgrids,
)
from ..utils import log_kaldi_errors

if TYPE_CHECKING:
    from logging import Logger

    from ..config import AlignConfig
    from ..corpus import Corpus
    from ..dictionary import DictionaryType
    from ..models import AcousticModel

__all__ = ["BaseAligner"]


class BaseAligner:
    """
    Base aligner class for common aligner functions

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus object for the dataset
    dictionary : :class:`~montreal_forced_aligner.dictionary.Dictionary`
        Dictionary object for the pronunciation dictionary
    align_config : :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Configuration for alignment
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    debug : bool
        Flag for running in debug mode, defaults to false
    verbose : bool
        Flag for running in verbose mode, defaults to false
    logger : :class:`~logging.Logger`
        Logger to use
    """

    def __init__(
        self,
        corpus: Corpus,
        dictionary: DictionaryType,
        align_config: AlignConfig,
        temp_directory: Optional[str] = None,
        debug: bool = False,
        verbose: bool = False,
        logger: Optional[Logger] = None,
        acoustic_model: Optional[AcousticModel] = None,
    ):
        self.align_config = align_config
        self.corpus = corpus
        self.dictionary = dictionary
        if not temp_directory:
            temp_directory = TEMP_DIR
        self.temp_directory = temp_directory
        os.makedirs(self.temp_directory, exist_ok=True)
        self.log_file = os.path.join(self.temp_directory, "aligner.log")
        if logger is None:
            self.logger = logging.getLogger("corpus_setup")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
        else:
            self.logger = logger
        self.acoustic_model = None
        self.verbose = verbose
        self.debug = debug
        self.speaker_independent = True
        self.uses_cmvn = True
        self.uses_splices = False
        self.uses_voiced = False
        self.iteration = None
        self.acoustic_model = acoustic_model
        self.setup()

    def setup(self) -> None:
        """
        Set up dictionary, corpus and configurations
        """
        self.dictionary.set_word_set(self.corpus.word_set)
        self.dictionary.write()
        self.corpus.initialize_corpus(self.dictionary, self.align_config.feature_config)
        self.align_config.silence_csl = self.dictionary.silence_csl
        self.data_directory = self.corpus.split_directory
        self.feature_config = self.align_config.feature_config

    @property
    def use_mp(self) -> bool:
        """Flag for using multiprocessing"""
        return self.align_config.use_mp

    @property
    def meta(self) -> dict:
        """Metadata for the trained model"""
        from ..utils import get_mfa_version

        data = {
            "phones": sorted(self.dictionary.nonsil_phones),
            "version": get_mfa_version(),
            "architecture": "gmm-hmm",
            "features": "mfcc+deltas",
        }
        return data

    @property
    def align_options(self):
        """Options for alignment"""
        options = self.align_config.align_options
        options["optional_silence_csl"] = self.dictionary.optional_silence_csl
        return options

    @property
    def fmllr_options(self):
        """Options for fMLLR"""
        options = self.align_config.fmllr_options
        options["silence_csl"] = self.dictionary.silence_csl
        return options

    @property
    def align_directory(self) -> str:
        """Align directory"""
        return os.path.join(self.temp_directory, "align")

    @property
    def working_directory(self) -> str:
        """Current working directory"""
        return self.align_directory

    @property
    def current_model_path(self) -> str:
        """Current acoustic model path"""
        return os.path.join(self.align_directory, "final.mdl")

    @property
    def alignment_model_path(self):
        """Alignment acoustic model path"""
        path = os.path.join(self.working_directory, "final.alimdl")
        if self.speaker_independent and os.path.exists(path):
            return path
        return os.path.join(self.working_directory, "final.mdl")

    @property
    def working_log_directory(self) -> str:
        """Current log directory"""
        return os.path.join(self.align_directory, "log")

    @property
    def backup_output_directory(self) -> Optional[str]:
        """Backup output directory"""
        if self.align_config.overwrite:
            return None
        return os.path.join(self.align_directory, "textgrids")

    def compile_information(self, output_directory: str) -> None:
        """
        Compile information about the quality of alignment

        Parameters
        ----------
        output_directory: str
            Directory to save information to
        """
        issues, average_log_like = compile_information(self)
        errors_path = os.path.join(output_directory, "output_errors.txt")
        if os.path.exists(errors_path):
            self.logger.warning(
                "There were errors when generating the textgrids. See the output_errors.txt in the "
                "output directory for more details."
            )
        if issues:
            issue_path = os.path.join(output_directory, "unaligned.txt")
            with open(issue_path, "w", encoding="utf8") as f:
                for u, r in sorted(issues.items()):
                    f.write(f"{u}\t{r}\n")
            self.logger.warning(
                f"There were {len(issues)} segments/files not aligned.  Please see {issue_path} for more details on why "
                "alignment failed for these files."
            )
        if (
            self.backup_output_directory is not None
            and os.path.exists(self.backup_output_directory)
            and os.listdir(self.backup_output_directory)
        ):
            self.logger.info(
                f"Some TextGrids were not output in the output directory to avoid overwriting existing files. "
                f"You can find them in {self.backup_output_directory}, and if you would like to disable this "
                f"behavior, you can rerun with the --overwrite flag or run `mfa configure --always_overwrite`."
            )

    def export_textgrids(self, output_directory: str) -> None:
        """
        Export a TextGrid file for every sound file in the dataset

        Parameters
        ----------
        output_directory: str
            Directory to save to
        """
        begin = time.time()
        self.textgrid_output = output_directory
        if self.backup_output_directory is not None and os.path.exists(
            self.backup_output_directory
        ):
            shutil.rmtree(self.backup_output_directory, ignore_errors=True)
        convert_ali_to_textgrids(self)
        self.compile_information(output_directory)
        self.logger.debug(f"Exported TextGrids in a total of {time.time() - begin} seconds")

    def align(self, subset: Optional[int] = None) -> None:
        """
        Perform alignment

        Parameters
        ----------
        subset: int, optional
            Number of utterances to align
        """
        done_path = os.path.join(self.align_directory, "done")
        dirty_path = os.path.join(self.align_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info("Alignment already done, skipping.")
            return
        try:
            compile_train_graphs(self)
            log_dir = os.path.join(self.align_directory, "log")
            os.makedirs(log_dir, exist_ok=True)

            self.logger.info("Performing first-pass alignment...")
            align(self)
            _, average_log_like = compile_information(self)
            self.logger.debug(
                f"Prior to SAT, average per frame likelihood (this might not actually mean anything): {average_log_like}"
            )
            if (
                not self.align_config.disable_sat
                and self.acoustic_model.feature_config.fmllr
                and not os.path.exists(os.path.join(self.align_directory, "trans.0"))
            ):
                self.logger.info("Calculating fMLLR for speaker adaptation...")
                calc_fmllr(self)
                self.logger.info("Performing second-pass alignment...")
                align(self)

                _, average_log_like = compile_information(self)
                self.logger.debug(
                    f"Following SAT, average per frame likelihood (this might not actually mean anything): {average_log_like}"
                )

        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        with open(done_path, "w"):
            pass



"""Excepts for Montreal Forced Aligner"""
from __future__ import annotations

from typing import TYPE_CHECKING, Collection, Dict, List, Optional, Tuple

from colorama import Fore, Style

from .helper import comma_join

if TYPE_CHECKING:
    from .dictionary import DictionaryType
    from .models import G2PModel


__all__ = [
    "MFAError",
    "SoxError",
    "G2PError",
    "ConfigError",
    "LMError",
    "LanguageModelNotFoundError",
    "ModelExtensionError",
    "ThirdpartyError",
    "TrainerError",
    "ModelError",
    "CorpusError",
    "ModelLoadError",
    "CorpusReadError",
    "ArgumentError",
    "AlignerError",
    "AlignmentError",
    "AlignmentExportError",
    "NoSuccessfulAlignments",
    "KaldiProcessingError",
    "TextParseError",
    "TextGridParseError",
    "DictionaryError",
    "NoDefaultSpeakerDictionaryError",
    "DictionaryPathError",
    "DictionaryFileError",
    "FileArgumentNotFoundError",
    "PretrainedModelNotFoundError",
    "MultipleModelTypesFoundError",
    "ModelTypeNotSupportedError",
    "PronunciationAcousticMismatchError",
    "PronunciationOrthographyMismatchError",
]


class MFAError(Exception):
    """
    Base exception class
    """

    def __init__(self, *args, **kwargs):
        from .config import USE_COLORS

        self.red_text = ""
        self.bright_text = ""
        self.green_text = ""
        self.reset_text = ""
        if USE_COLORS:
            self.red_text = Fore.RED
            self.bright_text = Style.BRIGHT
            self.green_text = Fore.GREEN
            self.reset_text = Style.RESET_ALL
        self.message = ""
        if args and isinstance(args[0], str):
            self.message = args[0]

    def error_text(self, text: str) -> str:
        """
        Highlight text as an error

        Parameters
        ----------
        text: str
            Text to highlight

        Returns
        -------
        str
            Highlighted text
        """
        return f"{self.red_text}{text}{self.reset_text}"

    def emphasized_text(self, text: str) -> str:
        """
        Highlight text as emphasis

        Parameters
        ----------
        text: str
            Text to highlight

        Returns
        -------
        str
            Highlighted text
        """
        return f"{self.bright_text}{text}{self.reset_text}"

    def pass_text(self, text: str) -> str:
        """
        Highlight text as good

        Parameters
        ----------
        text: str
            Text to highlight

        Returns
        -------
        str
            Highlighted text
        """
        return f"{self.green_text}{text}{self.reset_text}"

    def __str__(self) -> str:
        """Output the error"""
        return f"{self.error_text(type(self).__name__)}: {self.message}"


class ThirdpartyError(MFAError):
    """
    Exception class for errors in third party binary (usually Kaldi or OpenFst)

    Parameters
    ----------
    binary_name: str
        Name of third party binary
    open_fst: bool, optional
        Flag for the error having to do with OpenFst
    open_blas: bool, optional
        Flag for the error having to do with the BLAS library
    libc: bool, optional
        Flag for the error having to do with the system libraries
    sox: bool, optional
        Flag for the error having to do with SoX
    """

    def __init__(self, binary_name, open_fst=False, open_blas=False, libc=False, sox=False):
        super().__init__()
        self.message = f"Could not find '{self.error_text(binary_name)}'. "
        extra = "Please ensure that you have downloaded the correct binaries."
        if open_fst:
            extra = (
                f"Please ensure that you are in an environment that has the {self.emphasized_text('openfst')} conda package installed, "
                f"or that the {self.emphasized_text('openfst')} binaries are on your path if you compiled them yourself."
            )
        elif open_blas:
            extra = f"Try installing {self.emphasized_text('openblas')} via system package manager or verify it's on your system path?"
        elif libc:
            extra = (
                f"You likely have a different version of {self.emphasized_text('glibc')} than the precompiled binaries use. "
                f"Try compiling {self.emphasized_text('Kaldi')} on your machine and collecting the binaries via the "
                f"`{self.pass_text('mfa thirdparty kaldi')}` command."
            )
        elif sox:
            self.message = ""
            extra = (
                f"Your version of {self.emphasized_text('sox')} does not support the file format in your corpus. "
                f"Try installing another version of {self.emphasized_text('sox')} with support for {self.error_text(binary_name)}."
            )
        self.message += extra


# Model Errors


class ModelError(MFAError):
    """
    Exception class related to MFA model archives
    """

    pass


class ModelLoadError(ModelError):
    """
    Exception during loading of a model archive

    Parameters
    ----------
    path: str
        Path of the model archive
    """

    def __init__(self, path: str):
        super(ModelLoadError, self).__init__()
        self.message = f"The archive {self.error_text(path)} could not be parsed as an MFA model"


# Dictionary Errors


class DictionaryError(MFAError):
    """
    Exception class for errors in creating Dictionary objects
    """

    pass


class NoDefaultSpeakerDictionaryError(DictionaryError):
    """
    Exception class for errors in creating MultispeakerDictionary objects
    """

    def __init__(self):
        super().__init__()
        self.message = f'No "{self.error_text("default")}" dictionary was found.'


class DictionaryPathError(DictionaryError):
    """
    Exception class for errors in locating paths for Dictionary objects

    Parameters
    ----------
    input_path: str
        Path of the pronunciation dictionary
    """

    def __init__(self, input_path: str):
        super().__init__()
        self.message = (
            f"The specified path for the dictionary ({self.error_text(input_path)}) was not found."
        )


class DictionaryFileError(DictionaryError):
    """
    Exception class for file type being wrong for Dictionary objects

    Parameters
    ----------
    input_path: str
        Path of the pronunciation dictionary
    """

    def __init__(self, input_path: str):
        super().__init__()
        self.message = (
            f"The specified path for the dictionary ({self.error_text(input_path)}) is not a file."
        )


# Corpus Errors


class CorpusError(MFAError):
    """
    Class for errors in creating Corpus objects
    """

    pass


class CorpusReadError(CorpusError):
    """
    Class for errors in reading a file

    Parameters
    ----------
    file_name: str
        File name that was not readable
    """

    def __init__(self, file_name: str):
        MFAError.__init__(self)
        self.file_name = file_name


class TextParseError(CorpusReadError):
    """
    Class for errors parsing lab and txt files

    Parameters
    ----------
    file_name: str
        File name that had the error
    """

    def __init__(self, file_name: str):
        MFAError.__init__(self)
        self.file_name = file_name


class TextGridParseError(CorpusReadError):
    """
    Class capturing TextGrid reading errors

    Parameters
    ----------
    file_name: str
        File name that had the error
    error: str
        Error in TextGrid file
    """

    def __init__(self, file_name: str, error: str):
        MFAError.__init__(self)
        self.file_name = file_name
        self.error = error


class SoxError(CorpusReadError):
    """
    Class for errors in calling and finding Sox
    """

    pass


# Aligner Errors


class AlignerError(MFAError):
    """
    Class for errors during alignment
    """

    pass


class AlignmentError(MFAError):
    """
    Class for errors during alignment

    Parameters
    ----------
    error_logs: List[str]
        List of Kaldi log files with errors
    """

    def __init__(self, error_logs: List[str]):
        super().__init__()
        output = "\n".join(error_logs)
        self.message = (
            f"There were {len(error_logs)} job(s) with errors.  "
            f"For more information, please see the following logs:\n\n{output}"
        )


class AlignmentExportError(AlignmentError):
    """
    Class for errors in exporting alignments

    Parameters
    ----------
    error_dict: Dict[Tuple[str, int], str]
        Error dictionary mapping export stage and job to the error encountered

    """

    def __init__(self, error_dict: Dict[Tuple[str, int], str]):
        MFAError.__init__(self)

        message = "Error was encountered in processing CTMs:\n\n"
        for key, error in error_dict.items():
            message += f"{key}:\n{error}"
        self.message = message


class NoSuccessfulAlignments(AlignerError):
    """
    Class for errors where nothing could be aligned
    """

    pass


class PronunciationAcousticMismatchError(AlignerError):
    """
    Exception class for when an acoustic model and pronunciation dictionary have different phone sets

    Parameters
    ----------
    missing_phones: Collection[str]
        Phones that are not in the acoustic model
    """

    def __init__(self, missing_phones: Collection[str]):
        super().__init__()
        missing_phones = [f"{self.error_text(x)}" for x in sorted(missing_phones)]
        self.message = (
            f"There were phones in the dictionary that do not have acoustic models: "
            f"{comma_join(missing_phones)}"
        )


class PronunciationOrthographyMismatchError(AlignerError):
    """
    Exception class for missing graphemes in a G2P model

    Parameters
    ----------
    g2p_model: G2PModel
        Specified G2P model
    dictionary: Dictionary
        Specified dictionary
    """

    def __init__(self, g2p_model: G2PModel, dictionary: DictionaryType):
        super().__init__()
        missing_graphs = dictionary.graphemes - set(g2p_model.meta["graphemes"])
        missing_graphs = [f"{self.error_text(x)}" for x in sorted(missing_graphs)]
        self.message = (
            f"There were graphemes in the corpus that are not covered by the G2P model: "
            f"{comma_join(missing_graphs)}"
        )


# Command line exceptions


class ArgumentError(MFAError):
    """
    Exception class for errors parsing command line arguments
    """

    pass


class FileArgumentNotFoundError(ArgumentError):
    """
    Exception class for not finding a specified file

    Parameters
    ----------
    path: str
        Path not found
    """

    def __init__(self, path):
        super().__init__()
        self.message = f'Could not find "{self.error_text(path)}".'


class PretrainedModelNotFoundError(ArgumentError):
    """
    Exception class for not finding a specified pretrained model

    Parameters
    ----------
    name: str
        Model name
    model_type: str, optional
        Model type searched
    available: List[str], optional
        List of models that were found
    """

    def __init__(
        self, name: str, model_type: Optional[str] = None, available: Optional[List[str]] = None
    ):
        super().__init__()
        extra = ""
        if model_type:
            extra += f" for {model_type}"
        message = f'Could not find a model named "{self.error_text(name)}"{extra}.'
        if available:
            available = [f"{self.pass_text(x)}" for x in available]
            message += f" Available: {comma_join(available)}."
        self.message = message


class MultipleModelTypesFoundError(ArgumentError):
    """
    Exception class for finding multiple model types that could map to a given name

    Parameters
    ----------
    name: str
        Model name
    possible_model_types: List[str]
        List of model types that have a model with the given name
    """

    def __init__(self, name: str, possible_model_types: List[str]):
        super().__init__()

        possible_model_types = [f"{self.error_text(x)}" for x in possible_model_types]
        self.message = (
            f'Found multiple model types for "{self.error_text(name)}": {", ".join(possible_model_types)}. '
            f"Please specify a model type to inspect."
        )


class ModelExtensionError(ArgumentError):
    """
    Exception class for a model not having the correct extension

    Parameters
    ----------
    name: str
        Model name
    model_type: str
        Model type
    extensions: List[str]
        Extensions that the model supports
    """

    def __init__(self, name: str, model_type: str, extensions: List[str]):
        super().__init__()
        extra = ""
        if model_type:
            extra += f" for {model_type}"
        message = (
            f'The path "{self.error_text(name)}" does not have the correct extensions{extra}.'
        )
        if extensions:
            available = [f"{self.pass_text(x)}" for x in extensions]
            message += f" Possible extensions: {comma_join(available)}."
        self.message = message


class ModelTypeNotSupportedError(ArgumentError):
    """
    Exception class for a model type not being supported

    Parameters
    ----------
    model_type: str
        Model type
    model_types: List[str]
        List of supported model types
    """

    def __init__(self, model_type, model_types):
        super().__init__()
        message = f'The model type "{self.error_text(model_type)}" is not supported.'
        if model_types:
            model_types = [f"{self.pass_text(x)}" for x in sorted(model_types)]
            message += f" Possible model types: {comma_join(model_types)}."
        self.message = message


class ConfigError(MFAError):
    """
    Exception class for errors in configuration
    """

    pass


class TrainerError(MFAError):
    """
    Exception class for errors in trainers
    """

    pass


class G2PError(MFAError):
    """
    Exception class for errors in G2P
    """

    pass


class LMError(MFAError):
    """
    Exception class for errors in language models
    """

    pass


class LanguageModelNotFoundError(LMError):
    """
    Exception class for a language model not being found
    """

    def __init__(self):
        super(LanguageModelNotFoundError, self).__init__()
        self.message = "Could not find a suitable language model"


class KaldiProcessingError(MFAError):
    """
    Exception class for when a Kaldi binary has an exception

    Parameters
    ----------
    error_logs: List[str]
        List of Kaldi logs that had errors
    log_file: str, optional
        Overall log file to find more information
    """

    def __init__(self, error_logs: List[str], log_file: Optional[str] = None):
        super().__init__()
        self.message = (
            f"There were {len(error_logs)} job(s) with errors when running Kaldi binaries."
        )
        if log_file is not None:
            self.message += f" For more details, please check {self.error_text(log_file)}"
        self.error_logs = error_logs
        self.log_file = log_file

    def update_log_file(self, log_file: str) -> None:
        """
        Update the log file output

        Parameters
        ----------
        log_file: str
            Path to log file
        """
        self.log_file = log_file
        self.message = (
            f"There were {len(self.error_logs)} job(s) with errors when running Kaldi binaries."
        )
        if self.log_file is not None:
            self.message += f" For more details, please check {self.error_text(log_file)}"
